import itertools
import os
import pickle
import sys
from copy import copy
from ctypes import CFUNCTYPE, c_int64, POINTER, c_double, c_void_p

import numpy as np
import time
import uuid

from numba import njit, carray
import numpy.typing as npt
from numba.core.registry import CPUDispatcher

from numba.experimental import jitclass
import pandas as pd

from numerous.engine.model.events import generate_event_action_ast, generate_event_condition_ast, _replace_path_strings
from numerous.engine.model.utils import Imports, njit_and_compile_function
from numerous.engine.model.external_mappings import ExternalMapping, EmptyMapping
from numerous.engine.numerous_event import NumerousEvent, TimestampEvent

from numerous.utils.logger_levels import LoggerLevel

from numerous.utils.historian import InMemoryHistorian
from numerous.engine.model.graph_representation.mappings_graph import MappingsGraph
from numerous.engine.model.compiled_model import numba_model_spec, CompiledModel
from numerous.engine.system.connector import Connector

from numerous.engine.system.subsystem import Subsystem, ItemSet
from numerous.engine.variables import VariableType

from numerous.engine.model.ast_parser.parser_ast import parse_eq
from numerous.engine.model.graph_representation.graph import Graph
from numerous.engine.model.ast_parser.parser_ast import process_mappings

from numerous.engine.model.lowering.equations_generator import EquationGenerator
from numerous.engine.system import SetNamespace

import faulthandler
import llvmlite.binding as llvm

faulthandler.enable()
llvm.initialize()
llvm.initialize_native_target()
llvm.initialize_native_asmprinter()
llvmmodule = llvm.parse_assembly("")
target_machine = llvm.Target.from_default_triple().create_target_machine()
ee = llvm.create_mcjit_compiler(llvmmodule, target_machine)


class ModelNamespace:

    def __init__(self, tag, outgoing_mappings, item_tag, item_indcs, path, pos, item_path):
        self.tag = tag
        self.item_tag = item_tag
        self.outgoing_mappings = outgoing_mappings
        self.equation_dict = {}
        self.eq_variables_ids = []
        ##Ordered dictionary.
        self.variables = {}
        self.set_variables = None
        self.mappings = []
        self.full_tag = item_path + '.' + tag
        self.item_indcs = item_indcs

        self.path = path
        self.is_set = pos

    def ordered_variables(self):
        """
        return variables ordered for sequential llvm addressing
        """
        variables__ = []
        for vs in self.variable_scope:
            variables___ = []
            for v in vs:
                variables___.append(v)
            variables__.append(variables___)
        variables__ = [list(x) for x in zip(*variables__)]
        variables__ = list(itertools.chain(*variables__))
        return variables__


class ModelAssembler:

    @staticmethod
    def namespace_parser(input_namespace):
        equation_dict = {}
        tag, namespaces = input_namespace
        variables_ = {}
        for namespace in namespaces:
            for i, (eq_tag, eq_methods) in enumerate(namespace.equation_dict.items()):
                scope_id = "{0}_{1}_{2}".format(eq_tag, namespace.tag, tag, str(uuid.uuid4()))
                equation_dict.update({scope_id: (eq_methods, namespace.outgoing_mappings)})
            for v in namespace.ordered_variables():
                variables_.update({v.id: v})

        return variables_, equation_dict


import logging

logging.basicConfig(
    format='%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')


class Model:
    """
     The model object traverses the system to collect all information needed to pass to the solver
     for computation – the model also back-propagates the numerical results from the solver into the system,
     so they can be accessed as variable values there.
    """

    def __init__(self, system=None, logger_level=None, historian_filter=None, assemble=True, validate=False,
                 external_mappings=None, data_loader=None, imports=None, historian=InMemoryHistorian(),
                 use_llvm=True, save_to_file=False, generate_graph_pdf=False, export_model=False):

        self.path_to_variable = {}
        self.generate_graph_pdf = generate_graph_pdf
        self.export_model = export_model
        if logger_level == None:
            self.logger_level = LoggerLevel.ALL
        else:
            self.logger_level = logger_level

        self.is_external_data = True if external_mappings else False
        self.external_mappings = ExternalMapping(external_mappings,
                                                 data_loader) if external_mappings else EmptyMapping()
        self.use_llvm = use_llvm
        self.save_to_file = save_to_file
        self.imports = Imports()
        self.imports.add_as_import("numpy", "np")
        self.imports.add_from_import("numba", "njit")
        self.imports.add_from_import("numba", "carray")
        self.imports.add_from_import("numba", "float64")
        self.imports.add_from_import("numba", "float32")
        if imports:
            for (k, v) in imports:
                self.imports.add_from_import(k, v)

        self.numba_callbacks_init = []
        self.numba_callbacks_variables = []
        self.numba_callbacks = []
        self.numba_callbacks_init_run = []
        self.callbacks = []
        self.historian_filter = historian_filter
        self.system = system
        self.event_function = None
        self.condition_function = None
        self.derivatives = {}
        self.model_items = {}
        self.state_history = {}
        self.compiled_eq = []
        self.flat_scope_idx = None
        self.flat_scope_idx_from = None
        self.historian_df = None
        self.aliases = {}
        self.historian = historian
        self.vars_ordered_value = {}
        self.events = []
        self.timestamp_events = []
        self.global_variables_tags = ['time']
        self.global_vars = np.array([0], dtype=np.float64)

        self.equation_dict = {}
        self.variables = {}
        self.name_spaces = {}
        self.flat_variables = {}
        self.path_variables = {}
        self.states = {}
        self.period = 1
        self.mapping_from = []
        self.mapping_to = []
        self.eq_outgoing_mappings = []
        self.sum_mapping_from = []
        self.sum_mapping_to = []
        self.states_idx = []
        self.derivatives_idx = []
        self.scope_to_variables_idx = []
        self.numba_model = None

        self.info = {}
        if assemble:
            self.assemble()

        if validate:
            self.validate()

    def __add_item(self, item):
        model_namespaces = []
        if item.id in self.model_items:
            return model_namespaces

        self.model_items.update({item.id: item})
        model_namespaces.append((item.id, self.create_model_namespaces(item)))
        if isinstance(item, ItemSet):
            return model_namespaces
        if isinstance(item, Connector):
            for binded_item in item.get_binded_items():
                if not binded_item.part_of_set:
                    model_namespaces.extend(self.__add_item(binded_item))
        if isinstance(item, Subsystem):
            for registered_item in item.registered_items.values():
                model_namespaces.extend(self.__add_item(registered_item))
        return model_namespaces

    def __get_mapping__variable(self, variable, depth):
        if variable.mapping and depth > 0:
            return self.__get_mapping__variable(variable.mapping, depth - 1)
        else:
            return variable

    def assemble(self):
        """
        Assembles the model.
        """
        """
        notation:
        - _idx for single integers / tuples,
        - _idxs for lists / arrays of integers
        - _pos as counterpart to _from
        -  _flat
        -  _3d

        """

        def __get_mapping__idx(variable):
            if variable.mapping:
                return __get_mapping__idx(variable.mapping)
            else:
                return variable.idx_in_scope[0]

        logging.info("Assembling numerous Model")
        assemble_start = time.time()

        # 1. Create list of model namespaces
        model_namespaces = [_ns
                            for item in self.system.registered_items.values() if not item.part_of_set
                            for _ns in self.__add_item(item)]

        # 2. Compute dictionaries
        # equation_dict <scope_id, [Callable]>
        # scope_variables <variable_id, Variable>
        for variables, equation_dict in map(ModelAssembler.namespace_parser, model_namespaces):
            self.equation_dict.update(equation_dict)
            self.variables.update(variables)

        mappings = []
        for variable in self.variables.values():
            variable.top_item = self.system.id

        for scope_var_idx, var in enumerate(self.variables.values()):
            if var.mapping:
                _from = self.__get_mapping__variable(self.variables[var.mapping.id], depth=1)
                mappings.append((var.id, [_from.id]))
            if not var.mapping and var.sum_mapping:
                sum_mapping = []
                for mapping_id in var.sum_mapping:
                    _from = self.__get_mapping__variable(self.variables[mapping_id.id], depth=1)
                    sum_mapping.append(_from.id)
                mappings.append((var.id, sum_mapping))

        self.mappings_graph = Graph(preallocate_items=1000000)

        nodes_dep = {}
        self.equations_parsed = {}
        self.scoped_equations = {}
        self.equations_top = {}

        logging.info('Parsing equations starting')
        for v in self.variables.values():
            v.top_item = self.system.id

        eq_used = []
        for ns in model_namespaces:
            # will be false for empty namespaces. Ones without equations and variables.
            if ns[1]:
                # Key : scope.tag Value: Variable or VariableSet
                if ns[1][0].is_set:
                    tag_vars = ns[1][0].set_variables
                else:
                    tag_vars = {v.tag: v for k, v in ns[1][0].variables.items()}

                parse_eq(model_namespace=ns[1][0], item_id=ns[0], mappings_graph=self.mappings_graph,
                         scope_variables=tag_vars, parsed_eq_branches=self.equations_parsed,
                         scoped_equations=self.scoped_equations, parsed_eq=self.equations_top, eq_used=eq_used)
        self.eq_used = eq_used
        logging.info('Parsing equations completed')

        # Process mappings add update the global graph
        logging.info('Process mappings')
        self.mappings_graph = process_mappings(mappings, self.mappings_graph, self.variables)
        self.mappings_graph.build_node_edges()

        logging.info('Mappings processed')

        # Process variables
        states = []
        deriv = []
        mapping = []
        other = []

        for sv_id, sv in self.variables.items():
            if sv.logger_level is None:
                sv.logger_level = LoggerLevel.ALL
            if sv.type == VariableType.STATE:
                states.append(sv)
            elif sv.type == VariableType.DERIVATIVE:
                deriv.append(sv)
            elif sv.sum_mapping or sv.mapping:
                mapping.append(sv)
            else:
                other.append(sv)

        self.vars_ordered = states + deriv + mapping + other
        self.states_end_ix = len(states)

        self.deriv_end_ix = self.states_end_ix + len(deriv)
        self.mapping_end_ix = self.deriv_end_ix + len(mapping)

        self.special_indcs = [self.states_end_ix, self.deriv_end_ix, self.mapping_end_ix]

        logging.info('Variables sorted')

        self.mappings_graph = MappingsGraph.from_graph(self.mappings_graph)
        self.mappings_graph.remove_chains()
        tmp_vars = self.mappings_graph.create_assignments(self.variables)
        self.mappings_graph.add_mappings()
        if self.generate_graph_pdf:
            self.mappings_graph.as_graphviz(self.system.tag, force=True)
        self.lower_model_codegen(tmp_vars)
        self.assembly_tail()

    def assembly_tail(self):
        self.logged_aliases = {}

        for i, variable in enumerate(self.variables.values()):
            if variable.temporary_variable:
                continue
            if variable.logger_level is None:
                variable.logger_level = LoggerLevel.ALL
            logvar = False
            if variable.logger_level.value >= self.logger_level.value:
                logvar = True
            for path in variable.path.path[self.system.id]:
                self.aliases.update({path: variable.id})
                if logvar:
                    self.logged_aliases.update({path: variable.id})
            if variable.alias is not None:
                self.aliases.update({variable.alias: variable.id})
                if logvar:
                    self.logged_aliases.update({variable.alias: variable.id})
            for path in variable.path.path[self.system.id]:
                self.path_variables.update({path: variable.value})  # is this used at all?

        self.inverse_aliases = {v: k for k, v in self.aliases.items()}
        inverse_logged_aliases = {}  # {id: [alias1, alias2...], ...}
        for k, v in self.logged_aliases.items():
            inverse_logged_aliases[v] = inverse_logged_aliases.get(v, []) + [k]

        self.inverse_logged_aliases = inverse_logged_aliases
        self.logged_variables = {}

        for varname, ix in self.vars_ordered_values.items():  # now it's a dict...
            var = self.variables[varname]
            if var.logger_level.value >= self.logger_level.value:
                if varname in self.inverse_logged_aliases:
                    for vv in self.inverse_logged_aliases[varname]:
                        self.logged_variables.update({vv: ix})

        number_of_external_mappings = 0
        external_idx = []

        for var in self.variables.values():
            if self.external_mappings.is_mapped_var(self.variables, var.id, self.system.id):
                external_idx.append(self.vars_ordered_values[var.id])
                number_of_external_mappings += 1
                self.external_mappings.add_df_idx(self.variables, var.id, self.system.id)

        self.number_of_external_mappings = number_of_external_mappings
        self.external_mappings.store_mappings()
        self.external_idx = np.array(external_idx, dtype=np.int64)
        self.generate_path_to_varaible()

        # check for item level events
        for item in self.model_items.values():
            if item.events:
                for event in item.events:
                    event.condition = _replace_path_strings(self, event.condition, "state", item.path)
                    event.action = _replace_path_strings(self, event.action, "var", item.path)
                    self.events.append(event)
            if item.timestamp_events:
                for event in item.timestamp_events:
                    event.action = _replace_path_strings(self, event.action, "var", item.path)
                    self.timestamp_events.append(event)

        self.info.update({"Number of items": len(self.model_items)})
        self.info.update({"Number of variables": len(self.variables)})
        self.info.update({"Number of equation scopes": len(self.equation_dict)})
        self.info.update({"Number of equations": len(self.compiled_eq)})
        self.info.update({"Solver": {}})

    def lower_model_codegen(self, tmp_vars):

        logging.info('Lowering model')

        eq_gen = EquationGenerator(equations=self.equations_parsed, filename="kernel.py",
                                   equation_graph=self.mappings_graph,
                                   scope_variables=self.variables, scoped_equations=self.scoped_equations,
                                   temporary_variables=tmp_vars, system_tag=self.system.tag, use_llvm=self.use_llvm,
                                   imports=self.imports, eq_used=self.eq_used)

        compiled_compute, var_func, var_write, self.vars_ordered_values, self.variables, \
        self.state_idx, self.derivatives_idx = \
            eq_gen.generate_equations(export_model=self.export_model)

        for varname, ix in self.vars_ordered_values.items():
            var = self.variables[varname]
            var.llvm_idx = ix
            if getattr(var, 'logger_level',
                       None) is None:  # added to temporary variables - maybe put in generate_equations?
                setattr(var, 'logger_level', LoggerLevel.ALL)

        def c1(self, array_):
            return compiled_compute(array_)

        def c2(self):
            return var_func()

        def c3(self, value, idx):
            return var_write(value, idx)

        setattr(CompiledModel, "compiled_compute", c1)
        setattr(CompiledModel, "read_variables", c2)
        setattr(CompiledModel, "write_variables", c3)

        self.compiled_compute, self.var_func, self.var_write = compiled_compute, var_func, var_write
        self.init_values = np.ascontiguousarray(
            [self.variables[k].value for k in self.vars_ordered_values.keys()],
            dtype=np.float64)
        self.update_all_variables()

        # values of all model variables in specific order: self.vars_ordered_values
        # full tags of all variables in the model in specific order: self.vars_ordered
        # dict with scope variable id as key and scope variable itself as value

        # Create aliases for all paths in each scope variable
        def c4(values_dict):
            return [self.var_write(v, self.vars_ordered_values[self.aliases[k]]) for k, v in values_dict.items()]

        def c5():
            vals = var_func()
            return {self.inverse_aliases[k]: v for k, v in zip(self.vars_ordered_values.keys(), vals)}

        setattr(self, "update_variables", c4)
        setattr(self, "get_variables", c5)

        self.info.update({"Solver": {}})
        if self.export_model:
            path = os.environ.get("EXPORT_MODEL_PATH", 'export_model')
            filename = os.path.join(path, f'{self.system.tag}.numerous')
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, 'wb') as handle:
                sys.setrecursionlimit(100000)
                pickle.dump((self.system, self.logger_level, self.external_mappings,
                             self.imports, self.use_llvm, self.vars_ordered_values, self.variables,
                             self.state_idx, self.derivatives_idx, self.init_values, self.aliases,
                             eq_gen.generated_program.equations_llvm_opt, eq_gen.generated_program.max_var,
                             eq_gen.generated_program.n_deriv), handle, protocol=pickle.HIGHEST_PROTOCOL)
            logging.info('Model successfully exported to ' + filename)

        # set after export (otherwise we get pickling error for var.write_variable=var_write)
        for varname, ix in self.vars_ordered_values.items():
            var = self.variables[varname]
            var.llvm_idx = ix
            var.write_variable = var_write
            if getattr(var, 'logger_level',
                       None) is None:  # added to temporary variables - maybe put in generate_equations?
                setattr(var, 'logger_level', LoggerLevel.ALL)

        logging.info("Lowering model finished")

    def generate_path_to_varaible(self):
        for k, v in self.aliases.items():
            self.path_to_variable[k] = self.variables[v]

    def update_all_variables(self):
        for k, v in self.vars_ordered_values.items():
            self.var_write(self.variables[k].value, v)

    def update_local_variables(self):
        vars = self.numba_model.read_variables()
        for k, v in self.vars_ordered_values.items():
            self.variables[k].value = vars[v]

    def get_states(self):
        """

        Returns
        -------
        states : list of states
            list of all states.
        """
        return self.variables[self.state_idx]

    def update_states(self, y):
        self.variables[self.states_idx] = y

    def history_as_dataframe(self):
        time = self.data[0]
        data = {'time': time}

        for i, var in enumerate(self.var_list):
            data.update({var: self.data[i + 1]})

        self.df = pd.DataFrame(data)
        self.df = self.df.dropna(subset=['time'])
        self.df = self.df.set_index('time')
        self.df.index = pd.to_timedelta(self.df.index, unit='s')

    def validate_bindings(self):
        """
        Checks that all bindings are fulfilled.
        """
        valid = True
        for item in self.model_items.values():
            for binding in item.bindings:
                if binding.is_bindend():
                    pass
                else:
                    valid = False
        return valid

    def search_items(self, item_tag):
        """
        Search an item in items registered in the model by a tag

        Returns
        ----------
        items : list of :class:`numerous.engine.system.Item`
            set of items with given tag
               """
        return [item for item in self.model_items.values() if item.tag == item_tag]

    @property
    def states_as_vector(self):
        """
        Returns current states values.

        Returns
        -------
        state_values : array of state values

        """
        return self.var_func()[self.state_idx]

    def get_variable_path(self, id, item):
        for (variable, namespace) in item.get_variables():
            if variable.id == id:
                return "{0}.{1}".format(namespace.tag, variable.tag)
        if hasattr(item, 'registered_items'):
            for registered_item in item.registered_items.values():
                result = self.get_variable_path(id, registered_item)
                if result:
                    return "{0}.{1}".format(registered_item.tag, result)
        return ""

    def add_event(self, key, condition, action, terminal=True, direction=-1, compiled=False):
        condition = _replace_path_strings(self, condition, "state")

        action = _replace_path_strings(self, action, "var")
        event = NumerousEvent(key, condition, action, compiled, terminal, direction)
        self.events.append(event)

    def add_timestamp_event(self, key, action, timestamps):
        action = _replace_path_strings(self, action, "var")
        event = TimestampEvent(key, action, timestamps)
        self.timestamp_events.append(event)

    def generate_mock_event(self) -> None:
        def condition(t, v):
            return 1.0

        def action(t, v):
            i = 1

        self.add_event("mock", condition, action)

    def generate_mock_timestamp_event(self):
        def action(t, v):
            i = 1
        self.add_timestamp_event("mock", action, [-1])

    def generate_event_condition_ast(self, is_numerous_solver: bool) -> tuple[list[CPUDispatcher], npt.ArrayLike]:
        if len(self.events) == 0 and is_numerous_solver:
            self.generate_mock_event()
        if is_numerous_solver:
            return generate_event_condition_ast(self.events, self.imports.from_imports)
        else:
            result = []
            directions = []
            for event in self.events:
                if event.compiled:
                    compiled_event = event
                else:
                    compiled_event = njit_and_compile_function(event.condition, self.imports.from_imports)
                compiled_event.terminal = event.terminal
                compiled_event.direction = event.direction
                directions.append(event.direction)
                result.append(compiled_event)
            return result, np.array(directions)

    def generate_event_action_ast(self, events, is_numerous_solver: bool) -> list[CPUDispatcher]:
        if is_numerous_solver:
            return [generate_event_action_ast(events, self.imports.from_imports)]
        else:
            result = []
            for event in events:
                compiled_event = njit_and_compile_function(event.action, self.imports.from_imports)
                result.append(compiled_event)
            return result

    def _get_var_idx(self, var, idx_type):
        if idx_type == "state":
            return np.where(self.state_idx == var.llvm_idx)[0]
        if idx_type == "var":
            return [var.llvm_idx]

    def create_alias(self, variable_name, alias):
        """

        Parameters
        ----------
        variable_name
        alias

        Returns
        -------

        """
        self.variables[variable_name].alias = alias

    def create_model_namespaces(self, item):
        namespaces_list = []
        for namespace in item.registered_namespaces.values():
            set_namespace = isinstance(namespace, SetNamespace)
            model_namespace = ModelNamespace(namespace.tag, namespace.outgoing_mappings, item.tag, namespace.items,
                                             namespace.path, set_namespace, '.'.join(item.path))
            model_namespace.variable_scope = namespace.get_flat_variables()
            model_namespace.set_variables = namespace.set_variables

            equation_dict = {}
            eq_variables_ids = []
            for eq in namespace.associated_equations.values():
                equations = []
                ids = []
                for equation in eq.equations:
                    equations.append(equation)
                for vardesc in eq.variables_descriptions:
                    if set_namespace:
                        for item in namespace.items:
                            variable = item.registered_namespaces[namespace.tag].get_variable(vardesc.tag)
                            ids.append(variable.id)
                    else:
                        variable = namespace.get_variable(vardesc.tag)
                        ids.append(variable.id)
                equation_dict.update({eq.tag: equations})
                eq_variables_ids.append(ids)
            model_namespace.equation_dict = equation_dict
            model_namespace.variables = {v.id: v for vs in model_namespace.variable_scope for v in vs}
            namespaces_list.append(model_namespace)
        return namespaces_list

    # Method that generates numba_model
    def generate_compiled_model(self, start_time, number_of_timesteps):
        for spec_dict in self.numba_callbacks_variables:
            for item in spec_dict.items():
                numba_model_spec.append(item)

        # Creating a copy of CompiledModel class so it is possible
        # to creat instance detached from muttable type of CompiledModel
        tmp = type(f'{CompiledModel.__name__}' + self.system.id, CompiledModel.__bases__, dict(CompiledModel.__dict__))
        if self.use_llvm:
            @jitclass(numba_model_spec)
            class CompiledModel_instance(tmp):
                pass
        else:
            class CompiledModel_instance(tmp):
                pass

        NM_instance = CompiledModel_instance(self.init_values, self.derivatives_idx, self.state_idx,
                                             self.global_vars, number_of_timesteps, start_time,
                                             self.historian.get_historian_max_size(number_of_timesteps,
                                                                                   len(self.events)),
                                             self.external_mappings.external_mappings_time,
                                             self.number_of_external_mappings,
                                             self.external_mappings.external_mappings_numpy,
                                             self.external_mappings.external_df_idx,
                                             self.external_mappings.interpolation_info,
                                             self.is_external_data, self.external_mappings.t_max,
                                             self.external_idx
                                             )

        for key, value in self.path_variables.items():
            NM_instance.path_variables[key] = value
            NM_instance.path_keys.append(key)
        # NM_instance.run_init_callbacks(start_time)
        NM_instance.map_external_data(start_time)

        # NM_instance.historian_update(start_time)
        self.numba_model = NM_instance
        return self.numba_model

    def set_functions(self, compiled_compute, var_func, var_write):

        for varname, ix in self.vars_ordered_values.items():
            var = self.variables[varname]
            var.write_variable = var_write
            if getattr(var, 'logger_level',
                       None) is None:  # added to temporary variables - maybe put in generate_equations?
                setattr(var, 'logger_level', LoggerLevel.ALL)

        self.compiled_compute, self.var_func, self.var_write = compiled_compute, var_func, var_write

        def c4(values_dict):
            return [self.var_write(v, self.vars_ordered_values[self.aliases[k]]) for k, v in values_dict.items()]

        def c5():
            vals = var_func()
            return {self.inverse_aliases[k]: v for k, v in zip(self.vars_ordered_values.keys(), vals)}

        setattr(self, "update_variables", c4)
        setattr(self, "get_variables", c5)
        self.assembly_tail()

    def create_historian_dict(self, historian_data=None):
        if historian_data is None:
            historian_data = self.numba_model.historian_data
        time = historian_data[0]

        data = {var: historian_data[i + 1] for var, i in self.logged_variables.items()}
        data.update({'time': time})
        return data

    def generate_not_nan_history_array(self):
        return self.numba_model.historian_data[:, ~np.isnan(self.numba_model.historian_data).any(axis=0)]

    def create_historian_df(self):
        if self.historian_df is not None:
            import pandas as pd
            self.historian_df = AliasedDataFrame(pd.concat([self.historian_df,
                                                            self._generate_history_df(
                                                                self.generate_not_nan_history_array(),
                                                                rename_columns=False)],
                                                           axis=0, sort=False, ignore_index=True),
                                                 aliases=self.aliases, rename_columns=True)

        else:
            self.historian_df = self._generate_history_df(self.generate_not_nan_history_array(),
                                                          rename_columns=False)
        self.historian.store(self.historian_df)

    def _generate_history_df(self, historian_data, rename_columns=True):
        data = self.create_historian_dict(historian_data)
        return AliasedDataFrame(data, aliases=self.aliases, rename_columns=True)

    @classmethod
    def from_file(cls, param):
        system_, logger_level, external_mappings, imports, use_llvm, vars_ordered_values, variables, state_idx, \
        derivatives_idx, init_values, aliases, equations_llvm_opt, max_var, n_deriv = pickle.load(open(param, "rb"))
        if isinstance(external_mappings, EmptyMapping):
            external_mappings = None
        model = Model(system=system_, assemble=False, logger_level=logger_level, external_mappings=external_mappings,
                      use_llvm=use_llvm)
        model.variables = variables
        model.imports = imports
        model.vars_ordered_values = vars_ordered_values
        model.state_idx = state_idx
        model.derivatives_idx = derivatives_idx
        model.init_values = init_values
        model.aliases = aliases
        for equation in equations_llvm_opt:
            llmod2 = llvm.parse_assembly(equation)
            ee.add_module(llmod2)
        ee.finalize_object()
        cfptr = ee.get_function_address("kernel")

        cfptr_var_r = ee.get_function_address("vars_r")
        cfptr_var_w = ee.get_function_address("vars_w")

        c_float_type = type(np.ctypeslib.as_ctypes(np.float64()))

        diff_ = CFUNCTYPE(POINTER(c_float_type), POINTER(c_float_type))(cfptr)

        vars_r = CFUNCTYPE(POINTER(c_float_type), c_int64)(cfptr_var_r)

        vars_w = CFUNCTYPE(c_void_p, c_double, c_int64)(cfptr_var_w)

        @njit('float64[:](float64[:])')
        def compiled_compute(y):
            deriv_pointer = diff_(y.ctypes)
            return carray(deriv_pointer, (n_deriv,)).copy()

        @njit('float64[:]()')
        def var_func():
            variables_pointer = vars_r(0)
            variables_array = carray(variables_pointer, (max_var,))

            return variables_array.copy()

        @njit('void(float64,int64)')
        def var_write(var, idx):
            vars_w(var, idx)

        def c1(self, array_):
            return compiled_compute(array_)

        def c2(self):
            return var_func()

        def c3(self, value, idx):
            return var_write(value, idx)

        setattr(CompiledModel, "compiled_compute", c1)
        setattr(CompiledModel, "read_variables", c2)
        setattr(CompiledModel, "write_variables", c3)

        model.set_functions(compiled_compute, var_func, var_write)
        return model


class AliasedDataFrame(pd.DataFrame):
    _metadata = ['aliases']

    def __init__(self, data, aliases={}, rename_columns=True):
        super().__init__(data)
        self.aliases = aliases
        self.rename_columns = rename_columns
        if self.rename_columns:
            tmp = copy(list(data.keys()))
            for key in tmp:
                if key in self.aliases.keys():
                    data[self.aliases[key]] = data.pop(key)
            super().__init__(data)

    def __getitem__(self, item):
        if self.rename_columns:
            if not isinstance(item, list):
                col = self.aliases[item] if item in self.aliases else item
                return super().__getitem__(col)

            cols = [self.aliases[i] if i in self.aliases else i for i in item]

            return super().__getitem__(cols)
        else:
            return super().__getitem__(item)



import uuid
from collections import deque
import numpy as np
import numbers

from numerous.engine.variables import VariableBase, VariableDescription, VariableType
from numerous.multiphysics.equation_decorators import InlineEquation


class VariableDescriptionMap(VariableBase):
    def __init__(self, equation):
        self.variables_descriptions = super().__dict__
        self.equation = equation
        self.variables_descriptions_deque = deque()

    def variable_exists(self, tag):
        return tag in self.variables_descriptions

    def register_variable_description(self, variable_description):
        if variable_description.tag not in self.variables_descriptions:
            self.variables_descriptions[variable_description.tag] = variable_description
        else:
            self.variables_descriptions[variable_description.tag].update = True
            return
            raise ValueError('Variable description with tag {} is already exist in equation {}'.format(
                variable_description.tag, self.equation.tag))

    # refactored in a more functional way
    def __iter__(self):
        # used filter and extend
        self.variables_descriptions_deque.extend(list(filter((lambda v:\
            isinstance(v, VariableDescription)),\
            self.variables_descriptions.values())))
        return self

    def __next__(self):
        if self.variables_descriptions_deque:
            return self.variables_descriptions_deque.pop()
        else:
            raise StopIteration

class Mapping:
    def __init__(self, from_, to_=None, dir='<-'):
        self.from_ = from_
        self.to_ = to_ if to_ is not None else from_
        self.dir = dir


class EquationBase:
    """
    Equations governing the behavior of the objects in the model and their interactions
     are defined in classes extending the :class:`numerous.multiphysics.Equation`.

    """
    def __init__(self, tag=None):
        if tag:
            self.tag = tag
        self.equations = []
        self.variables_descriptions = VariableDescriptionMap(self)
        super(EquationBase, self).__init__()
        method_list = [func for func in dir(self) if callable(getattr(self, func)) and not func.startswith("__")]
        for method in method_list:
            method_call = getattr(self, method)
            if hasattr(method_call, '_equation'):
                self.equations.append(method_call)


    def add_parameter(self, tag, init_val=0, logger_level=None, alias=None, integrate=None):
        """

        Parameters
        ----------
        tag
        init_val

        Returns
        -------

        """
        self.add_variable(tag, init_val, VariableType.PARAMETER, logger_level, alias)

        if integrate is not None:
            self.add_state(integrate['tag'], 0, logger_level=logger_level)

            integrate_source = f"""
            def integrate_{tag}(self, scope):
                scope.{integrate['tag']}_dot = scope.{tag} * {integrate['scale']}
            """
            ie = InlineEquation()
            integrate_source_ = ie('integrate_'+tag, integrate_source, {})

            setattr(self, 'integrate_'+tag, integrate_source_)
            self.equations.append(integrate_source_)


    def add_parameters(self, parameters:dict or list):
        if isinstance(parameters, dict):
            for p, v in parameters.items():
                if isinstance(v,numbers.Number):
                    self.add_parameter(p, init_val=v)
                else:
                    self.add_parameter(p, **v)
        if isinstance(parameters, list):
            for p in parameters:
                self.add_parameter(p)

    def add_constants(self, constants:dict):
        for p, v in constants.items():
            if isinstance(v,numbers.Number):
                self.add_constant(p, value=v)
            else:
                self.add_constant(p, **v)

    def add_constant(self, tag, value, logger_level=None, alias=None):
        """

        Parameters
        ----------
        tag
        value

        Returns
        -------

        """
        self.add_variable(tag, value, VariableType.CONSTANT, logger_level, alias)

    def add_state(self, tag, init_val=0, logger_level=None, alias=None):
        """

        Parameters
        ----------
        tag
        init_val

        Returns
        -------

        """
        if not isinstance(init_val, float) and not isinstance(init_val, int) and not isinstance(init_val, np.int64):
            raise ValueError("State must be float or integer")
        self.add_variable(tag, init_val, VariableType.STATE, logger_level, alias)
        self.add_variable(tag + '_dot', 0, VariableType.DERIVATIVE, logger_level,
                          alias+"_dot" if alias is not None else None)

    def add_variable(self, tag, init_val, var_type, logger_level, alias):
        """

        Parameters
        ----------
        tag
        init_val
        var_type
        external_mappng

        Returns
        -------

        """
        self.variables_descriptions. \
            register_variable_description(VariableDescription(tag=tag, id=str(uuid.uuid1()), initial_value=init_val,
                                                              type=var_type, logger_level=logger_level, alias=alias))

    def map_create_parameters(self, item, mappings):
        for m in mappings:
            if not self.variables_descriptions.variable_exists(m.to_):
                self.add_parameter(m.to_)






import time

from scipy.integrate import solve_ivp
from tqdm import tqdm
import numpy as np

from numerous.engine.simulation.solvers.base_solver import BaseSolver

"""
Wrapper for scipy ivp solver.
"""


class IVP_solver(BaseSolver):

    def __init__(self, time, delta_t, model, numba_model, num_inner, max_event_steps, y0, events, timestamp_events, **kwargs):
        super().__init__()
        self.model = model
        self.time = time
        self.eventf = []
        self.event_a = []
        for event_cond in events[0]:
            self.eventf.append(event_cond)
        for event_action in events[1]:
            self.event_a.append(event_action)

        self.timestamps_actions = timestamp_events[0]
        self.timestamps = timestamp_events[1]
        self.y0 = y0
        self.num_inner = num_inner
        self.delta_t = delta_t
        self.numba_model = numba_model
        self.diff_function = numba_model.func
        self.max_event_steps = max_event_steps
        self.options = kwargs
        self.t = 0

    def solve(self):
        """
        solve the model.
        Returns
        -------
        Solution : 'OdeSolution'
                returns the most recent OdeSolution from scipy
        """
        self.result_status = "Success"
        self.sol = None

        try:
            for t in tqdm(self.time[0:-1]):
                if self.solver_step(t):
                    break
        except Exception as e:
            print(e)
            raise e

        return self.sol, self.result_status

    def solver_step(self, t, delta_t=None):
        step_not_finished = True
        current_timestamp = t
        event_steps = 0

        step_solver_mode = False
        if delta_t is None:
            delta_t = self.delta_t
        else:
            step_solver_mode = True

        stop_condition = False
        t_end = current_timestamp + delta_t

        while step_not_finished:
            t_eval = np.linspace(current_timestamp, t_end, self.num_inner + 1)

            self.sol = solve_ivp(self.diff_function, (current_timestamp, t + delta_t), y0=self.y0, t_eval=t_eval,
                                 dense_output=False, events=self.eventf,
                                 **self.options)
            step_not_finished = False
            event_step = self.sol.status == 1

            if self.sol.status == 0:
                current_timestamp = t_end
                if step_solver_mode:  # added this
                    self.model.numba_model.historian_update(current_timestamp)
                    self.y0 = self.sol.y[:, -1]
                    return current_timestamp, self.sol.t[-1]
                for id, timestamps in enumerate(self.timestamps):
                    for timestamp in timestamps:
                        if abs(timestamp - current_timestamp) < 1e-6:
                            self.__end_step(self, self.sol.y, current_timestamp, self.timestamps_actions, event_id=id)
            if event_step:
                event_id = np.nonzero([x.size > 0 for x in self.sol.t_events])[0][0]
                # solution stuck
                stop_condition = False
                if (abs(self.sol.t_events[event_id][0] - current_timestamp) < 1e-6):
                    event_steps += 1
                else:
                    event_steps = 0

                if event_steps > self.max_event_steps:
                    stop_condition = True
                current_timestamp = self.sol.t_events[event_id][0]
                step_not_finished = True
                self.__end_step(self, self.sol.y, current_timestamp,self.event_a, event_id=event_id)
            else:
                if self.sol.success:
                    self.__end_step(self, self.sol.y[:, -1], current_timestamp,self.event_a)
                else:
                    self.result_status = self.sol.message
            if stop_condition:
                break

        self.t = t_end
        if stop_condition:
            self.result_status = "Stopping condition reached"
            return True
        return False

    def set_state_vector(self, states_as_vector):
        self.y0 = states_as_vector

    def register_endstep(self, __end_step):
        self.__end_step = __end_step



import ast
from enum import unique, IntEnum

from numba.core.registry import CPUDispatcher


class Imports:
    def __init__(self):
        self.as_imports = []
        self.from_imports = []

    def add_as_import(self, module_name, alias):
        self.as_imports.append((module_name, alias))

    def add_from_import(self, module_name, element):
        self.from_imports.append((module_name, element))


@unique
class EventTypes(IntEnum):
    TIME_EVENT = 0
    STATE_EVENT = 1


# TODO add setVAR/GridVar?
@unique
class NodeTypes(IntEnum):
    OP = 0
    ASSIGN = 1
    VAR = 2
    EQUATION = 3
    SUM = 4
    TMP = 5
    ASSIGNTUPLE = 6
    IF = 7


def recurse_Attribute(attr, sep='.'):
    if hasattr(attr, 'id'):
        return attr.id
    elif isinstance(attr.value, ast.Name):
        return attr.value.id + sep + attr.attr
    elif isinstance(attr.value, ast.Attribute):
        return recurse_Attribute(attr.value) + sep + attr.attr


def njit_and_compile_function(func: ast.FunctionDef, from_imports: list[(str, str)]) -> CPUDispatcher:
    fname = func.name
    njit_decorator = ast.Name(id='njit', ctx=ast.Load())
    func.decorator_list = [njit_decorator]
    body = []
    for (module, label) in from_imports:
        body.append(
            ast.ImportFrom(module=module, names=[ast.alias(name=label, asname=None)], lineno=0, col_offset=0,
                           level=0))
    body.append(ast.Import(names=[ast.alias(name='numpy', asname='np')]))
    body.append(func)
    body.append(
        ast.Return(value=ast.Name(id=fname, ctx=ast.Load(), lineno=0, col_offset=0), lineno=0, col_offset=0))

    func = wrap_function(fname + '1', body, decorators=[],
                         args=ast.arguments(posonlyargs=[], args=[], vararg=None, defaults=[],
                                            kwonlyargs=[], kw_defaults=[], lineno=0, kwarg=None))
    module_func = ast.Module(body=[func], type_ignores=[])
    code = compile(ast.parse(ast.unparse(module_func)), filename='event_storage', mode='exec')
    namespace = {}
    exec(code, namespace)
    compiled_func = list(namespace.values())[1]()
    return compiled_func


def wrap_function(name, body, args, decorators):
    f = ast.FunctionDef(name)
    f.body = body
    f.decorator_list = decorators
    f.args = args
    f.lineno = 0
    f.col_offset = 0
    return f



import uuid


class Node:
    """
        Top element in an items hierarchy. Contains fields to represent as
        uniques modelling object.


       Attributes
       ----------
       tag :  string
            Not unique tag that will be used in reports or printed output.
       id : string
            UUID assigned to the node. Generated if not provided.

    """

    def __init__(self, tag=None, id_=None):
        if not isinstance(tag, str):
            raise TypeError(f'Tag must be a string. Cannot be {type(tag)}')
        if tag:
            self.tag = tag
        else:
            self.tag = 'not_tagged'
        if id_ is None:
            self.id = self.tag + str(uuid.uuid1())
        else:
            if self.id is None:
                self.id = self.tag + uuid.UUID(id_)
        super(Node, self).__init__()


    @property
    def get_id(self):
        """
        Returns
        -------
        id : string
            UUID of the element.
        """
        return self.id

    def __repr__(self):
        return self.tag



import numpy as np

from numerous.engine.model.external_mappings.interpolation_type import InterpolationType


class ExternalMapping:
    def __init__(self, external_mappings, data_loader):
        self.external_mappings = external_mappings
        self.data_loader = data_loader
        self.external_mappings_numpy = []
        self.external_mappings_time = []
        self.external_columns = []
        self.interpoaltion_type = []
        self.t_max = 0
        # loading initial dataframes
        for element in self.external_mappings:
            df = self.data_loader.load(element.data_frame_id, element.index_to_timestep_mapping_start)
            element.add_df(df)
        # mapping data
        for element in self.external_mappings:
            self.external_columns.append(list(element.dataframe_aliases.keys()))
            self.interpoaltion_type.append([a_tuple[1] for a_tuple in list(element.dataframe_aliases.values())])
            ext_mapping_numpy = element.df[[a_tuple[0] for a_tuple in list(element.dataframe_aliases.values())]]
            self.external_mappings_numpy.append(
                ext_mapping_numpy.to_numpy(dtype=np.float64)[element.index_to_timestep_mapping_start:]
            )
            self.external_mappings_time.append(
                element.time_multiplier * element.df[element.index_to_timestep_mapping].to_numpy(dtype=np.float64)[
                                          element.index_to_timestep_mapping_start:])
            # TODO extend for multiple dataframes
            self.t_max = np.max(self.external_mappings_time[0])
        self.external_mappings_numpy = np.array(self.external_mappings_numpy, dtype=np.float64)
        self.external_mappings_time = np.array(self.external_mappings_time, dtype=np.float64)
        self.interpoaltion_type = [item for sublist in self.interpoaltion_type for item in sublist]
        self.external_df_idx = []
        self.interpolation_info = []

    def load_new_external_data_batch(self, t):
        self.external_mappings_numpy = []
        self.external_mappings_time = []

        for element in self.external_mappings:
            # TODO division round bugs? we can skip a row here
            df = self.data_loader.load(element.data_frame_id, int(t / element.time_multiplier) - 1)

            if df.empty:
                return False

            element.add_df(df)
        for element in self.external_mappings:
            self.external_mappings_numpy.append(
                element.df[
                    [a_tuple[0] for a_tuple in list(element.dataframe_aliases.values())]
                ].to_numpy(dtype=np.float64)[element.index_to_timestep_mapping_start:])
            self.external_mappings_time.append(
                element.time_multiplier * element.df[element.index_to_timestep_mapping].to_numpy(dtype=np.float64)[
                                          element.index_to_timestep_mapping_start:])
        self.external_mappings_numpy = np.array(self.external_mappings_numpy, dtype=np.float64)
        self.external_mappings_time = np.array(self.external_mappings_time, dtype=np.float64)
        # TODO extend for multiple dataframes
        self.t_max = np.max(self.external_mappings_time[0])
        return True

    def store_mappings(self):
        self.external_df_idx = np.array(self.external_df_idx, dtype=np.int64)
        self.interpolation_info = np.array(self.interpolation_info, dtype=np.bool)

    def add_df_idx(self, variables, var_id, system_id):
        for i, external_column in enumerate(self.external_columns):
            for path in variables[var_id].path.path[system_id]:
                if path in external_column:
                    i1 = external_column.index(path)
                    self.external_df_idx.append((i, i1))
                    self.interpolation_info.append(
                        self.interpoaltion_type[i].value == InterpolationType.LINEAR.value)

    def is_mapped_var(self, variables, var_id, system_id):
        for path in variables[var_id].path.path[system_id]:
            for columns in self.external_columns:
                if path in columns:
                    return True


class ExternalMappingElement:

    def __init__(self, data_frame_id, index_to_timestep_mapping, index_to_timestep_mapping_start, time_multiplier,
                 dataframe_aliases):
        self.data_frame_id = data_frame_id
        self.index_to_timestep_mapping = index_to_timestep_mapping
        self.index_to_timestep_mapping_start = index_to_timestep_mapping_start
        self.time_multiplier = time_multiplier
        self.dataframe_aliases = dataframe_aliases
        self.df = None

    # TODO check that e_m is correct format (have all the mappings) after loading
    def add_df(self, df):
        if self.df is not None:
            columns = self.df.columns
            self.df = df
            self.df.columns = columns
        else:
            self.df = df


class EmptyMapping:
    def __init__(self):
        self.external_mappings_numpy = np.empty([0, 0, 0], dtype=np.float64)
        self.external_mappings_time = np.empty([0, 0], dtype=np.float64)
        self.external_df_idx = np.empty([0, 0], dtype=np.int64)
        self.t_max = 0
        self.interpolation_info = np.empty([0], dtype=bool)

    def store_mappings(self):
        pass

    def is_mapped_var(self, variables, var_id, system_id):
        pass



from numba import jit
import numpy as np

@jit(nopython=True, nogil=True, cache=True)
def bilinear_interpolation(ix, x, y, data_def, data):
    '''Interpolate (x,y) from values associated with four points.

    The four points are a list of four triplets:  (x, y, value).
    The four points can be in any order.  They should form a rectangle.

        >>> bilinear_interpolation(12, 5.5,
        ...                        [(10, 4, 100),
        ...                         (20, 4, 200),
        ...                         (10, 6, 150),
        ...                         (20, 6, 300)])
        165.0

    '''
    ds_ix = data_def[ix]
    x_start_ix = ds_ix + 6

    x_len = data[ds_ix + 0]

    x_min = data[ds_ix + 2]

    x_max = data[ds_ix + 3]

    y_len = data[ds_ix + 1]

    y_min = data[ds_ix + 4]
    y_max = data[ds_ix + 5]

    y_start_ix = x_start_ix + np.int64(x_len)
    if (y_start_ix) < ds_ix:
        error_code = 3

    if x_max == x_min:
        x_norm = x
    else:
        x_norm = (x - x_min) / (x_max - x_min)

    if y_max == y_min:
        y_norm = y
    else:
        y_norm = (y - y_min) / (y_max - y_min)

    x_frac_ix = x_norm * (x_len - 1)

    x1_ix = np.int64(np.floor(x_frac_ix))

    if x1_ix < 0:
        x1_ix = 0
    elif x1_ix >= x_len:
        x1_ix = np.int64(x_len) - 1

    x1 = data[x1_ix + x_start_ix]

    x2_ix = np.int64(np.ceil(x_frac_ix))

    if x2_ix < 0:
        x2_ix = 0
    elif x2_ix >= x_len:
        x2_ix = np.int64(x_len) - 1

    x2 = data[x2_ix + x_start_ix]

    y_frac_ix = y_norm * (y_len - 1)

    y1_ix = np.int64(np.floor(y_frac_ix))

    if y1_ix < 0:
        y1_ix = 0
    elif y1_ix >= y_len:
        y1_ix = np.int64(y_len) - 1

    y1 = data[y1_ix + y_start_ix]

    y2_ix = np.int64(np.ceil(y_frac_ix))

    if y2_ix < 0:
        y2_ix = 0
    elif y2_ix >= y_len:
        y2_ix = np.int64(y_len) - 1

    y2 = data[y2_ix + y_start_ix]

    q_start_ix = y_start_ix + np.int64(y_len)

    q11 = data[q_start_ix + x1_ix + np.int64(x_len) * y1_ix]
    q12 = data[q_start_ix + x1_ix + np.int64(x_len) * y2_ix]
    q21 = data[q_start_ix + x2_ix + np.int64(x_len) * y1_ix]
    q22 = data[q_start_ix + x2_ix + np.int64(x_len) * y2_ix]

    # See formula at:  http://en.wikipedia.org/wiki/Bilinear_interpolation

    # points = sorted(points)               # order points by x, then by y
    # (x1, y1, q11), (_x1, y2, q12), (x2, _y1, q21), (_x2, _y2, q22) = points

    #    #   if x1 != _x1 or x2 != _x2 or y1 != _y1 or y2 != _y2:
    #     raise ValueError('points do not form a rectangle')
    # if not x1 <= x <= x2 or not y1 <= y <= y2:
    #     raise ValueError('(x, y) not within the rectangle')

    if x == x2 or x == x1 or x1 == x2:
        if y == y1 or y == y2 or y1 == y2:
            return q11
        else:
            z = (q11 * (y2 - y) + q12 * (y - y1)) / (y2 - y1)
    elif y == y1 or y == y2 or y1 == y2:
        if x == x1 or x == x2 or x1 == x2:
            return q11
        else:
            z = (q11 * (x2 - x) + q21 * (x - x1)) / (x2 - x1)
    else:
        z = (q11 * (x2 - x) * (y2 - y) +
             q21 * (x - x1) * (y2 - y) +
             q12 * (x2 - x) * (y - y1) +
             q22 * (x - x1) * (y - y1)
             ) / ((x2 - x1) * (y2 - y1))
    return z



import os

class Historian:
    def __init__(self, max_size, timesteps_reserved_per_event=100):
        self.max_size = max_size
        self.timesteps_reserved_per_event = timesteps_reserved_per_event

    def get_historian_max_size(self, number_of_timesteps, events_count):
        if not self.max_size:
            ## +1 here since we are using <= to compare inside compiled model
            return number_of_timesteps + events_count * self.timesteps_reserved_per_event + 1
        else:
            return self.max_size

    def store(self, df):
        pass



class InMemoryHistorian(Historian):

    def __init__(self):
        super().__init__(None)


class LocalHistorian(Historian):

    def __init__(self, filename, max_size):
        super().__init__(max_size)
        self.filename = filename

    def store(self, df):
        if os.path.isfile(self.filename):
            df.dropna().to_csv(self.filename, mode='a', header=False)
        else:
            df.dropna().to_csv(self.filename)


import logging
from copy import copy

import numpy as np
from graphviz import Digraph
from .utils import EdgeType, TemporaryKeyGenerator
from .lower_graph import multi_replace, _Graph

tmp_generator = TemporaryKeyGenerator().generate


class Node:
    def __init__(self, key=None, ao=None, file=None, name=None, ln=None,
                 label=None, ast_type=None, vectorized=None, item_id=None, node_type=None, ops=None, func=None,
                 id=None, local_id=None, scope_var=None,
                 ast_op=None, value=None, is_set_var=None, set_var_ix=None, subgraph_test=None,
                 subgraph_body=None):
        self.key = key
        self.ao = ao
        self.file = file
        self.value = value
        self.name = name
        self.ln = ln
        self.label = label
        self.ast_type = ast_type
        self.node_type = node_type
        self.ast_op = ast_op
        self.id = id
        self.vectorized = vectorized
        self.item_id = item_id
        self.is_set_var = is_set_var
        self.ops = ops
        self.set_var_ix = set_var_ix
        self.func = func
        self.local_id = local_id
        self.scope_var = scope_var
        self.deleted = False
        self.subgraph_test = subgraph_test
        self.subgraph_body = subgraph_body
        self.node_n = -1


class Edge:
    def __init__(self, start=-1, end=-1, e_type=EdgeType.UNDEFINED, label=None, arg_local=None, mappings=None,
                 branches=None):
        self.start = start
        self.end = end
        self.e_type = e_type
        self.arg_local = arg_local
        self.branches = branches
        self.label = label
        self.mappings = mappings
        self.deleted = False
        self.edge_n = -1


class Graph:

    def __init__(self, label=None, preallocate_items=1000):

        self.preallocate_items = preallocate_items
        self.allocated = self.preallocate_items
        self.edge_counter = 0
        self.node_counter = 0
        # Maps a key to an integer which is the internal node_id
        self.node_map = {}
        self.key_map = {}
        self.nodes = []
        self.edges_c = []
        self.label = label
        self.edges = np.ones((self.preallocate_items, 2), dtype=np.int32) * -1
        self.lower_graph = None

        self.node_edges = None

        ##For equation arguments order.
        self.arg_metadata = []
        self.skipped_arg_metadata = []

    def build_node_edges(self):
        self.node_edges = [([], []) for _ in range(self.node_counter)]
        for i, e in enumerate(self.edges[:self.edge_counter]):
            self.node_edges[e[0]][0].append(i)
            self.node_edges[e[1]][1].append(i)

    def add_node(self, node, ignore_existing=False, skip_existing=True):
        if not node.key:
            node.key = tmp_generator()
        if node.key not in self.node_map or ignore_existing:
            if not node.key in self.node_map:

                node_n = self.node_counter
                self.node_map[node.key] = node_n
                self.key_map[node_n] = node.key
                self.node_counter += 1

                if self.node_counter > self.allocated:
                    raise ValueError('Exceeding allocation')

            else:
                node_n = self.node_map[node.key]
            node.node_n = node_n
            if node_n < len(self.nodes):
                self.nodes[node_n] = node
            else:
                self.nodes.append(node)

        else:
            if not skip_existing:
                raise ValueError(f'Node with key already in graph <{node.key}>')
            else:
                return self.node_map[node.key]
        return node_n

    def add_edge(self, edge: Edge):
        edge_n = self.edge_counter
        self.edges[edge_n, :] = [edge.start, edge.end]

        self.edge_counter += 1

        if self.edge_counter > self.allocated:
            raise ValueError('Exceeding allocation')
        self.edges_c.append(edge)
        edge.edge_n = edge_n
        if edge.e_type == EdgeType.TARGET:
            pass
        return edge_n

    def set_edge(self, edge, start=None, end=None):
        if start:
            self.edges[edge, 0] = start
        if end:
            self.edges[edge, 1] = end

    def remove_node(self, node_n):
        self.nodes[node_n].deleted = True

    def clean(self):
        logging.info('Cleaning eq graph')
        self.lower_graph = None
        self.node_edges = None
        return self

    def remove_edge(self, edge_n):
        self.edges_c[edge_n].deleted = True

    def get(self, node, attr):
        return getattr(self.nodes[node], attr)

    def set(self, node, attr, val):
        setattr(self.nodes[node], attr, val)

    def get_where_node_attr(self, attr, val, not_=False):
        def filter_function(node):
            if node.deleted:
                return False
            if getattr(node, attr) == val:
                return True and not not_
            else:
                return False or not_

        return [node.node_n for node in filter(filter_function, self.nodes)]

    def get_edges_for_node(self, start_node=None, end_node=None):
        if start_node is not None:
            start_ix = self.edges[:self.edge_counter, 0] == start_node
            start_ = self.edges[:self.edge_counter][start_ix]
        else:
            start_ = self.edges[:self.edge_counter, :]
            start_ix = range(self.edge_counter)
        if end_node is not None:
            end_ix = start_[:, 1] == end_node
            end_ = self.edges[:self.edge_counter][end_ix]

        else:
            end_ix = start_ix
            end_ = start_

        return zip(np.argwhere(end_ix), end_)

    def get_edges_for_node_filter(self, attr, start_node: int = None, end_node: int = None, val=None):
        if start_node and end_node:
            raise ValueError('arg cant have both start and end!')
        ix = []
        if not self.node_edges:
            self.build_node_edges()

        if not start_node is None:
            ix = self.node_edges[start_node][0]

        if not end_node is None:
            ix = self.node_edges[end_node][1]

        if start_node is None and end_node is None:
            print(end_node)
            print(start_node)
            raise ValueError('Need at least one node!')

        def filter_function(edge):
            if edge.deleted:
                return False
            if getattr(edge, attr) == val :
                return True
            else:
                return False

        ix_r = [edge.edge_n for edge in filter(filter_function, map(self.edges_c.__getitem__, ix))]
        return ix_r, [self.edges[i, :] for i in ix_r]

    def has_edge_for_nodes(self, start_node=None, end_node=None):

        if start_node and end_node:
            return np.where(
                (self.edges[:self.edge_counter, 0] == start_node) & (self.edges[:self.edge_counter, 1] == end_node))[0]

        if start_node:
            return start_node in self.edges[:, 0]

        if end_node:
            return end_node in self.edges[:, 1]

    def clone(self):
        from copy import deepcopy
        clone_ = Graph(preallocate_items=self.preallocate_items)

        clone_.preallocate_items = self.preallocate_items
        clone_.edge_counter = self.edge_counter
        clone_.node_counter = self.node_counter
        # Maps a key to an integer which is the internal node_id
        clone_.node_map = self.node_map.copy()
        clone_.key_map = self.key_map.copy()

        clone_.nodes = self.nodes.copy()
        clone_.edges_c = deepcopy(self.edges_c)
        clone_.edges = self.edges.copy()

        return clone_

    def subgraph(self, nodes, edges):
        subgraph = Graph()
        sg_map = {}
        for n in nodes:
            node_copy = copy(self.nodes[n])
            node_copy.key = self.key_map[n]
            sg_map[n] = subgraph.add_node(node_copy)

        for e in edges:
            subgraph.add_edge(Edge(start=sg_map[e[0]], end=sg_map[e[1]]))
        return subgraph

    def as_graphviz(self, file, force=False):
        if False or force:
            dot = Digraph()
            for k, n in self.node_map.items():
                dot.node(k, label=self.nodes[n].label)

            for i, e in enumerate(self.edges[:self.edge_counter]):
                if not self.edges_c[i].deleted:
                    try:
                        if e[0] >= 0 and e[1] >= 0:
                            dot.edge(self.key_map[e[0]], self.key_map[e[1]], label=str(self.edges_c[i].e_type))
                    except Exception as e:
                        print(e)
                        raise

            dot.render(file, view=True, format='pdf')

    def zero_in_degree(self):
        return [n for n in self.node_map.values() if len(list(self.get_edges_for_node(end_node=n))) == 0]

    def make_lower_graph(self, top_sort=False):
        self.lower_graph = _Graph(self.node_counter,
                                  np.array(self.edges[:self.edge_counter], np.int64),
                                  np.array([node.node_type for node in self.nodes[:self.node_counter]], np.int64))

        if top_sort:
            self.lower_graph.topological_sort()

    def graph_from_path(self, path):
        cg = Graph()
        prev = None
        for p in path:
            new_node = Node(key=self.key_map[p], label=self.key_map[p])
            this_ = cg.add_node(new_node, ignore_existing=True)
            if prev is not None:
                cg.add_edge(Edge(start=prev, end=this_, e_type=EdgeType.DEP))
            prev = this_

        return cg

    def topological_nodes(self, ignore_cyclic=True):
        logging.info('Starting topological sort')
        if not self.lower_graph:
            self.make_lower_graph()

        self.lower_graph.topological_sort()

        if self.lower_graph.cyclic_dependency >= 0:

            unsorted_nodes = set(self.lower_graph.nodes).difference(set(self.lower_graph.topological_sorted_nodes))
            self.cyclic_path = self.lower_graph.cyclic_path
            if not ignore_cyclic:
                cg = self.graph_from_path(self.cyclic_path)
                cg.as_graphviz('cyclic', force=True)
                for n in self.cyclic_path:
                    print(" ".join([str(self.key_map[n]), '          ' + str(
                        self.get(n, 'file'))]))

                self.cyclic_dependency = self.lower_graph.cyclic_dependency
                raise ValueError('Cyclic path detected: ', self.cyclic_path)
            else:
                self.lower_graph.topological_sorted_nodes[-len(unsorted_nodes):] = list(unsorted_nodes)

        logging.info("Topological sort finished")
        return self.lower_graph.topological_sorted_nodes

    def get_dependants_graph(self, node):
        if not self.lower_graph:
            self.make_lower_graph()
        nodes, edges = self.lower_graph.get_dependants_graph(np.array([node], np.int64))
        return self.subgraph(nodes, edges)

    def replace_nodes_by_key(self, key, to_be_replaced):
        n = self.node_map[key]
        to_be_replaced_v = np.array([self.node_map[k] for k in to_be_replaced], np.int64)
        edges = self.edges[:self.edge_counter]

        multi_replace(edges, to_be_replaced_v, n)
        [self.remove_node(t) for t in to_be_replaced_v]



class NumerousEvent:
    def __init__(self, key, condition, action, compiled, terminal, direction):
        self.key = key
        self.condition = condition
        self.action = action
        self.compiled = compiled
        self.direction = direction
        self.terminal = terminal


class TimestampEvent:
    def __init__(self, key, action, timestamps):
        self.key = key
        self.action = action
        self.timestamps = timestamps



import logging
from enum import IntEnum, unique

import ast


@unique
class VarTypes(IntEnum):
    CONSTANT = 0
    PARAMETER = 1
    DERIV = 2
    STATE = 3
    LOCAL = 4
    TMP = 5


def wrap_module(body):
    mod = ast.Module()
    for i in body:
        if not hasattr(i, "_fields"):
            setattr(i, "_fields", [])
    mod.body = body
    mod.type_ignores=[]
    setattr(mod, "_fields", [])
    return mod


def generate_code_file(mod_body, file, imports, external_functions_source=False, names="#"):
    for (module, name) in imports.as_imports:
        mod_body.insert(0, ast.Import(names=[ast.alias(name=module, asname=name)], level=0))
    for (module, name) in imports.from_imports:
        mod_body.insert(0, ast.ImportFrom(module=module, names=[ast.alias(name=name, asname=None)], level=0))
    if external_functions_source:
        mod_body.insert(0, ast.ImportFrom(module=external_functions_source, names=[ast.alias(name='*', asname=None)],
                                          level=0))

    mod = wrap_module(mod_body)
    logging.info('Generating Source')
    source = names + ast.unparse(mod)
    return source


# Define helper functions to check the variables and sub lists
def non_unique_check(listname_, list_):
    if len(list_) > len(set(list_)):
        import collections
        raise ValueError(
            f'Non unique {listname_}: {[item for item, count in collections.Counter(list_).items() if count > 1]}')


def contains_dot(str_list):
    for s in str_list:
        if '.' in s:
            raise ValueError(f'. in {s}')


def are_all_scalars(str_list, scalar_variables):
    for s in str_list:
        if not s in scalar_variables:
            raise ValueError(f'Not a scalar: {s}')


def are_all_set_variables(str_list, set_variables):
    for s in str_list:
        if not s in set_variables:
            raise ValueError(f'Not a set variable: {s}')


class Vardef:
    def __init__(self, llvm=True):
        self.vars_inds_map = []
        self.targets = []
        self.args = []
        self.llvm = llvm
        self.args_order = []
        self.trgs_order = []

    def format(self, var, read=True):
        if read:
            _ctx = ast.Load()
        else:
            _ctx = ast.Store()
        return ast.Name(id=var.replace('scope.', 's_'), lineno=0, col_offset=0, ctx=_ctx)

    def format_target(self, var,read):
        if read:
            _ctx = ast.Load()
        else:
            _ctx = ast.Store()
        if self.llvm:
            return ast.Subscript(slice=ast.Index(value=ast.Num(n=0, lineno=0,col_offset=0), lineno=0,col_offset=0), value=ast.Call(
                args=[ast.Name(id=var.replace('scope.', 's_'), lineno=0,col_offset=0, ctx=ast.Load()), ast.Tuple(ctx=ast.Load(),elts=[ast.Num(n=1,lineno=0,col_offset=0)], lineno=0,col_offset=0)],
                func=ast.Name(id='carray', lineno=0,col_offset=0, ctx=ast.Load()),
                keywords=[], lineno=0,col_offset=0), lineno=0,col_offset=0, ctx=_ctx)
        else:
            return ast.Name(id=var.replace('scope.', 's_'), lineno=0,col_offset=0, ctx=_ctx)

    def order_variables(self, order_data):
        for (var, var_id, used) in order_data:
            if used:
                self.args_order.append("scope." + var)
            else:
                self.args_order.append(var_id)

        for (var, var_id, used) in order_data:
            tmp_v = "scope." + var
            if tmp_v in self.targets:
                self.trgs_order.append(tmp_v)

    def var_def(self, var,ctxread, read=True):
        if not var in self.vars_inds_map:
            self.vars_inds_map.append(var)
        if read and 'scope.' in var:
            if var not in self.targets and var not in self.args:
                self.args.append(var)
        elif 'scope.' in var:

            if var not in self.targets:
                self.targets.append(var)

        if var in self.targets:
            return self.format_target(var, ctxread)
        else:
            return self.format(var, ctxread)

    def get_order_args(self, form=True):
        if form:
            result = [self.format(a,False) for a in self.args_order]
        else:
            result = self.args
        result_2 = []
        for name in result:
            result_2.append(ast.arg(arg=name.id,  lineno=0, col_offset=0))
        return result_2

    def get_order_trgs(self, form=True):
        if form:
            return [self.format(a,False) for a in self.trgs_order]
        else:
            return self.args

    def get_args(self, form=True):
        if form:
            return [self.format(a,False) for a in self.args]
        else:
            return self.args

    def get_targets(self, form=True):
        if form:
            return [self.format(a,False) for a in self.targets]
        else:
            return self.targets



import numba.extending as nbe
import ctypes
import numpy as np
from numba import njit

dpotrf_addr = nbe.get_cython_function_address('scipy.linalg.cython_lapack', 'dpotrf') # cholesky decomposition
dtrsv_addr = nbe.get_cython_function_address('scipy.linalg.cython_blas', 'dtrsv') # solve Ax=b

dpotrf_cfunc = ctypes.CFUNCTYPE(ctypes.c_int,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p)

dtrsv_cfunc = ctypes.CFUNCTYPE(ctypes.c_int,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p,
                                ctypes.c_void_p)

dpotrf_fun = dpotrf_cfunc(dpotrf_addr)
dtrsv_fun = dtrsv_cfunc(dtrsv_addr)

@njit
def lapack_solve_triangular(Lalloc, balloc, N):
    #solves the linalg equation A*x=L*L**T*x=b, where L is the lower cholesky decomposed matrix of A

    side_a = np.empty(1, dtype=np.int32)
    side_a[0] = 76  # L
    t_a = np.empty(1, dtype=np.int32)
    t_a[0] = 78 # N - solve L*x=b and not L**T*x=b
    diag_a = np.empty(1, dtype=np.int32)
    diag_a[0] = 78
    N_a = np.empty(1, dtype=np.int32)
    N_a[0] = N
    lda_a = np.empty(1, dtype=np.int32)
    lda_a[0] = N
    incx_a = np.empty(1, dtype=np.int32)
    incx_a[0] = 1

    # forward substitution using Lower diagonal matrix L*y = b, solves for y
    dtrsv_fun(side_a.ctypes,
              t_a.ctypes,
              diag_a.ctypes,
              N_a.ctypes,
              Lalloc.ctypes,
              lda_a.ctypes,
              balloc.ctypes,
              incx_a.ctypes
              )

    #side_a[0] = 85 # U
    t_a[0] = 84 # T - since the matrix L**T is the upper matrix
    #t_a[0] = 78 # T

    # backward substitution using upper diagonal matrix U*x=y, solves for x
    dtrsv_fun(side_a.ctypes,
              t_a.ctypes,
              diag_a.ctypes,
              N_a.ctypes,
              Lalloc.ctypes,
              lda_a.ctypes,
              balloc.ctypes,
              incx_a.ctypes
              )

    return balloc


@njit
def lapack_cholesky(side, N, xalloc):

    N_a = np.empty(1, dtype=np.int32)
    N_a[0] = N
    side_a = np.empty(1, dtype=np.int32)
    side_a[0] = side
    z_a = np.empty(1, dtype=np.int32)
    z_a[0] = 1

    dpotrf_fun(side_a.ctypes, N_a.ctypes, xalloc.ctypes, N_a.ctypes, z_a.ctypes)

    # return xalloc, containing the Lower cholesky decomposed matrix of xalloc stored in fortran order.
    # To get the actual lower matrix, transpose xalloc, but know that this slows down all subsequent code for some
    # reason

    return xalloc


import logging
import os
import sys
import time
import types

import numpy as np

from numerous.engine.model.ast_parser.equation_form_graph import function_from_graph_generic, \
    compiled_function_from_graph_generic_llvm
from numerous.engine.model.lowering.ast_builder import ASTBuilder
from numerous.engine.model.graph_representation import EdgeType

from numerous.engine.model.lowering.llvm_builder import LLVMBuilder
from numerous.engine.model.lowering.utils import Vardef
from numerous.engine.model.utils import NodeTypes, recurse_Attribute
from numerous.engine.variables import VariableType
from numerous.utils.string_utils import d_u


class EquationGenerator:
    def __init__(self, filename, equation_graph, scope_variables, equations, scoped_equations,
                 temporary_variables, system_tag="", use_llvm=True, imports=None, eq_used=[]):
        self.filename = filename
        self.imports = imports
        self.system_tag = system_tag
        self.scope_variables = scope_variables
        self.set_variables = {}
        self.states = []
        self.llvm_names = {}
        self.deriv = []

        for ix, (sv_id, sv) in enumerate(self.scope_variables.items()):
            full_tag = d_u(sv.id)
            if sv.type == VariableType.STATE:
                self.states.append(full_tag)
            elif sv.type == VariableType.DERIVATIVE:
                self.deriv.append(full_tag)

        for k, var in temporary_variables.items():
            if var.type == VariableType.TMP_PARAMETER_SET:
                self.set_variables.update({k: var})
                new_sv = {}
                tail = {}
                for k, v in self.scope_variables.items():
                    if k in var.set_var.variables:
                        tail.update({k: v})
                        new_sv.update({var.tmp_vars[v.set_var_ix].id: var.tmp_vars[v.set_var_ix]})
                    else:
                        new_sv.update({k: v})
                self.scope_variables = dict(new_sv, **tail)
            if var.type == VariableType.TMP_PARAMETER:
                new_sv = {}
                tail = {}
                for k, v in self.scope_variables.items():
                    if k == var.scope_var_id:
                        tail.update({k: v})
                        new_sv.update({var.id: var})
                    else:
                        new_sv.update({k: v})
                self.scope_variables = dict(new_sv, **tail)

        self.scoped_equations = scoped_equations
        self.temporary_variables = temporary_variables

        self.values_order = {}

        self.scope_var_node = {}
        self.scalar_variables = {}

        self._parse_variables()

        # Sort the graph topologically to start generating code
        self.topo_sorted_nodes = equation_graph.topological_nodes()
        self.equation_graph = equation_graph.clean()

        self.number_of_states = len(self.states)
        self.number_of_derivatives = len(self.deriv)

        # Initialize llvm builder - will be a list of intermediate llvm instructions to be lowered in generate
        self.llvm = use_llvm
        if self.llvm:
            self.generated_program = LLVMBuilder(
                np.ascontiguousarray([x.value for x in self.scope_variables.values()], dtype=np.float64),
                self.values_order, self.states, self.deriv, system_tag=self.system_tag)
        else:
            self.generated_program = ASTBuilder(
                np.ascontiguousarray([x.value for x in self.scope_variables.values()], dtype=np.float64),
                self.values_order, self.states, self.deriv, system_tag=self.system_tag)

        self.mod_body = []
        # Create a kernel of assignments and calls

        self.eq_vardefs = {}
        # Loop over equation functions and generate code for each equation.

        used_eq = {}

        for eq_key, eq in equations.items():
            if eq_key in eq_used:
                used_eq[eq_key]=eq
        self._parse_equations(used_eq)

        self.all_targeted = []
        self.all_read = []
        self.all_targeted_set_vars = []
        self.all_read_set_vars = []

    def _parse_variable(self, full_tag, sv, sv_id):

        if full_tag not in self.scope_var_node:
            self.scope_var_node[full_tag] = sv

        # If a scope_variable is part of a set it should be referenced alone
        if sv.set_var:
            if not sv.set_var.id in self.set_variables:
                self.set_variables[sv.set_var.id] = sv.set_var
        else:
            self.scalar_variables[full_tag] = sv

    def _parse_variables(self):
        for ix, (sv_id, sv) in enumerate(self.scope_variables.items()):
            full_tag = d_u(sv.id)
            self.values_order[full_tag] = ix
            self._parse_variable(full_tag, sv, sv_id)
        for ix, (sv_id, sv) in enumerate(self.temporary_variables.items()):
            full_tag = d_u(sv.id)
            self._parse_variable(full_tag, sv, sv_id)

    def get_external_function_name(self, ext_func):
        if self.llvm:
            return self._llvm_func_name(ext_func)
        return ext_func

    def _llvm_func_name(self, ext_func):
        return self.llvm_names[ext_func + '_llvm1.<locals>.' + ext_func + '_llvm']

    def _parse_equations(self, equations):
        logging.info('Making equations for compilation')

        for eq_key, eq in equations.items():
            vardef = Vardef(llvm=self.llvm)

            eq.graph.lower_graph = None
            if self.llvm:
                func_llvm, signature, args, target_ids = compiled_function_from_graph_generic_llvm(
                    eq.graph,
                    imports=self.imports,
                    var_def_=Vardef(llvm=self.llvm),
                    compiled_function=True,
                    replacements=eq.replacements,
                    replace_name=eq_key
                )
                self.llvm_names.update(self.generated_program.add_external_function(func_llvm, signature, len(args), target_ids,
                                                             replacements=eq.replacements,replace_name=eq_key))

            else:
                func, args, target_ids = function_from_graph_generic(eq.graph,
                                                                     var_def_=vardef, arg_metadata=eq.graph.arg_metadata,
                                                                     replacements=eq.replacements,
                                                                     replace_name=eq_key
                                                                     )
                self.generated_program.add_external_function(func, None, len(args), target_ids,
                                                             replacements=eq.replacements,replace_name=eq_key)

            vardef.llvm_target_ids = target_ids
            vardef.args_order = args
            self.eq_vardefs[eq_key] = vardef

    def search_in_item_scope(self, var_id, item_id):
        for var in self.scope_variables.values():
            ##TODO add namespacecheck
            if var.item.id == item_id and var.tag == self.scope_variables[var_id].tag:
                return var.id
        raise ValueError("No variable found for id {}", var_id)

    def _process_equation_node(self, n):

        eq_key = self.scoped_equations[self.equation_graph.key_map[n]]

        # Define the function to call for this eq
        ext_func = recurse_Attribute(self.equation_graph.get(n, 'func'))
        item_id = self.equation_graph.get(n, 'item_id')

        vardef = self.eq_vardefs[eq_key]

        # Find the arguments by looking for edges of arg type
        a_indcs, a_edges = list(
            self.equation_graph.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.ARGUMENT))
        # Determine the local arguments names
        args_local = [self.equation_graph.key_map[ae[0]] for i, ae in zip(a_indcs, a_edges) if
                      not self.equation_graph.edges_c[i].arg_local == 'local']

        # Determine the local arguments names
        args_scope_var = [self.equation_graph.edges_c[i].arg_local for i, ae in zip(a_indcs, a_edges) if
                          not self.equation_graph.edges_c[i].arg_local == 'local']

        # Find the targets by looking for target edges
        t_indcs, t_edges = list(
            self.equation_graph.get_edges_for_node_filter(start_node=n, attr='e_type', val=EdgeType.TARGET))
        targets_local = [self.equation_graph.key_map[te[1]] for i, te in zip(t_indcs, t_edges) if
                         not self.equation_graph.edges_c[i].arg_local == 'local']
        targets_scope_var = [self.equation_graph.edges_c[i].arg_local for i, ae in zip(t_indcs, t_edges)
                             if
                             not self.equation_graph.edges_c[i].arg_local == 'local']
        set_size = 0
        # Record targeted and read variables
        if self.equation_graph.get(n, 'vectorized'):

            # Map of scope.?? vars and set variable names
            scope_vars = {'scope.' + self.set_variables[k].tag: v for k, v in
                          zip(args_scope_var + targets_scope_var, args_local + targets_local)}

            # Put the information of args and targets in the scope_var attr of the graph node for those equation
            self.equation_graph.nodes[n].scope_var = {'args': [scope_vars[a] for a in vardef.args],
                                                      'targets': [scope_vars[a] for a in vardef.targets]}
            # Record all targeted variables
            for t in vardef.targets:
                self.all_targeted_set_vars.append(scope_vars[t])
                ##TODO check that they all the same size
                set_size = self.set_variables[scope_vars[t]].get_size()
            # Record all read variables
            for a in vardef.args:
                self.all_read_set_vars.append(scope_vars[a])
        else:
            # Map of scope.?? vars and global-scope variable names
            scope_vars = {'scope.' + self.scope_variables[k].tag: v for k, v in
                          zip(args_scope_var + targets_scope_var, args_local + targets_local)}

            # Put the information of args and targets in the scope_var attr of the graph node for those equation
            self.equation_graph.nodes[n].scope_var = {'args': [scope_vars[a] for a in vardef.args],
                                                      'targets': [scope_vars[a] for a in vardef.targets]}
            for a in vardef.args:
                if (sva := scope_vars[a]) in self.set_variables:
                    self.all_read_set_vars.append(sva)
                else:
                    self.all_read.append(sva)

            self.all_targeted += [scope_vars[t] for t in
                                  vardef.targets]

        # Generate ast for this equation callcompiled_function_from_graph_generic_llvm
        if self.equation_graph.get(n, 'vectorized'):

            llvm_args = []
            for t in vardef.args_order:
                llvm_args_ = []
                set_var = self.set_variables[scope_vars[t]]
                for i in range(set_var.get_size()):
                    llvm_args_.append(set_var.get_var_by_idx(i).id)
                llvm_args.append(llvm_args_)
            ##reshape to correct format
            llvm_args = [list(x) for x in zip(*llvm_args)]
            self.generated_program.add_set_call(self.get_external_function_name(ext_func), llvm_args,
                                                vardef.llvm_target_ids)
        else:
            # Generate llvm arguments
            args = []

            for a in vardef.args_order:
                if a in scope_vars:
                    args.append(d_u(scope_vars[a]))
                else:
                    args.append(self.search_in_item_scope(a, item_id))

            # Add this eq to the llvm_program
            self.generated_program.add_call(self.get_external_function_name(ext_func), args, vardef.llvm_target_ids)

    def _process_sum_node(self, n):
        t_indcs, target_edges = list(
            self.equation_graph.get_edges_for_node_filter(start_node=n, attr='e_type', val=EdgeType.TARGET))
        v_indcs, value_edges = list(
            self.equation_graph.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.VALUE))

        # assume single target
        if lte := len(target_edges) != 1:
            raise ValueError(f'Wrong length of target edges - must be 1 but is {lte}')
        t = target_edges[0][1]

        # If the target is a set variable
        if (t_sv := self.equation_graph.get(t, 'scope_var')).size > 0:

            self.all_targeted_set_vars.append(self.equation_graph.key_map[t])

            l_mapping = t_sv.size
            mappings = {':': [], 'ix': []}

            # make a list of assignments to each index in t
            for v_ix, v in zip(v_indcs, value_edges):
                if (nt := self.equation_graph.get(v[0], 'node_type')) == NodeTypes.VAR or nt == NodeTypes.TMP:

                    if (mix := self.equation_graph.edges_c[v_ix].mappings) == ':':
                        mappings[':'].append(self.equation_graph.key_map[v[0]])

                    elif isinstance(mix, list):
                        sums = {}
                        for m in mix:
                            if not m[1] in sums:
                                sums[m[1]] = []
                            sums[m[1]].append(m[0])
                        mappings['ix'].append((self.equation_graph.key_map[v[0]], sums))

                    else:
                        raise ValueError(
                            f'mapping indices not specified!{self.equation_graph.edges_c[v_ix].mappings}, {self.equation_graph.key_map[t]} <- {self.equation_graph.key_map[v[0]]}')

                else:
                    raise ValueError(f'this must be a mistake {self.equation_graph.key_map[v[0]]}')

            mappings_ast_pairs = [
                                     []] * l_mapping  # To be list of tuples of var and target for each indix in target

            # process specific index mappings
            for m_ix in mappings['ix']:

                # m_ix[0] is a variable mapped to the current set variable
                # m_ix[1] is a dict:
                # Keys which are indices to this set variable.
                # Values which are indices to the variable mapped to this set variable
                from_ = m_ix[0]
                # m_ix1_keys = list(m_ix[1].keys())

                # loop over all indices in target
                for target_ix, value_ix in m_ix[1].items():
                    mappings_ast_pairs[target_ix].append((from_, value_ix))

            # Generate llvm
            mappings_llvm = {}
            if len(mappings[':']) > 0:
                for set_variable in mappings[':']:
                    for k in range(self.set_variables[set_variable].size):
                        target_var_name = self.set_variables[t_sv.id].get_var_by_idx(k).id
                        mappings_llvm[target_var_name] = [self.set_variables[set_variable].get_var_by_idx(k).id]
            for m_ix in mappings['ix']:
                for k, v in m_ix[1].items():
                    target_var_name = self.set_variables[t_sv.id].get_var_by_idx(k).id
                    if target_var_name not in mappings_llvm.keys():
                        mappings_llvm[target_var_name] = []
                    for el in v:
                        if m_ix[0] in self.set_variables:
                            mappings_llvm[target_var_name].append(self.set_variables[m_ix[0]].get_var_by_idx(el).id)
                        else:
                            if m_ix[0] in self.scope_variables:
                                mappings_llvm[target_var_name].append(self.scope_variables[m_ix[0]].id)
                            else:
                                raise ValueError(f'Variable  {m_ix[0]} mapping not found')
            for k, v in mappings_llvm.items():
                self.generated_program.add_mapping(v, [k])
        else:
            # Register targeted variables
            if is_set_var := self.equation_graph.get(t, attr='is_set_var'):
                self.all_targeted_set_vars.append(self.equation_graph.key_map[t])
            else:

                self.all_targeted.append(self.equation_graph.key_map[t])

            target_indcs_map = [[] for i in
                                range(len(
                                    self.set_variables[self.equation_graph.key_map[t]]))] if is_set_var else [
                []]

            for v, vi in zip(value_edges, v_indcs):
                if self.equation_graph.get(v[0], 'is_set_var'):
                    self.all_read_set_vars.append(self.equation_graph.key_map[v[0]])

                else:
                    self.all_read.append(self.equation_graph.key_map[v[0]])

                maps = self.equation_graph.edges_c[vi].mappings

                if maps == ':':
                    if self.equation_graph.key_map[t] in self.set_variables:
                        for mi in range(len(self.set_variables[self.equation_graph.key_map[t]])):
                            target_indcs_map[mi].append((v[0], mi))
                    else:
                        target_indcs_map[0].append((v[0], None))
                else:
                    for mi in maps:
                        target_indcs_map[mi[1] if mi[1] else 0].append((v[0], mi[0]))

            target_var = self.equation_graph.key_map[t]

            # Generate llvm/ast
            mapping_dict = {}
            for values in target_indcs_map:
                for v in values:
                    var_name = d_u(self.equation_graph.key_map[v[0]])
                    if var_name in self.scope_variables:
                        if target_var in mapping_dict:
                            mapping_dict[target_var].append(var_name)
                        else:
                            mapping_dict[target_var] = [var_name]
                    else:
                        if var_name in self.set_variables:
                            if var_name in mapping_dict:
                                mapping_dict[target_var].append(self.set_variables[var_name].get_var_by_idx(v[1]).id)
                            else:
                                mapping_dict[target_var] = [self.set_variables[var_name].get_var_by_idx(v[1]).id]
                        else:
                            raise ValueError(f'Variable  {var_name} mapping not found')
            for k, v in mapping_dict.items():
                self.generated_program.add_mapping(v, [k])

    def generate_equations(self, export_model=False):
        logging.info('Generate kernel')
        # Generate the ast for the python kernel
        for n in self.topo_sorted_nodes:
            # Add the equation calls
            if (nt := self.equation_graph.get(n, 'node_type')) == NodeTypes.EQUATION:
                self._process_equation_node(n)
            # # Add the sum statements
            elif nt == NodeTypes.SUM:
                self._process_sum_node(n)

            elif nt == NodeTypes.VAR or nt == NodeTypes.TMP:
                pass
            else:
                raise ValueError('Unused node: ', self.equation_graph.key_map[n])

        deriv_idx = []
        state_idx = []
        for k, v in self.values_order.items():
            if k in self.deriv:
                deriv_idx.append(v)
            if k in self.states:
                state_idx.append(v)
        if self.llvm:
            logging.info('Generating llvm')
            diff, var_func, var_write = self.generated_program.generate(imports=self.imports,
                                                                        system_tag=self.system_tag,
                                                                        save_to_file=export_model)

            return diff, var_func, var_write, self.values_order, self.scope_variables, np.array(state_idx,
                                                                                                dtype=np.int64), \
                   np.array(
                       deriv_idx, dtype=np.int64)
        else:

            global_kernel, var_func, var_write = self.generated_program.generate(self.imports,
                                                                                 system_tag=self.system_tag,
                                                                                 save_to_file=export_model)
            return global_kernel, var_func, var_write, self.values_order, self.scope_variables, \
                   np.array(state_idx, dtype=np.int64), np.array(deriv_idx, dtype=np.int64)


class ItemPath:
    """
           Represent a unique item path to item from the top of hierarchy..

          Attributes
          ----------
          path :  string
            path to item in string form
          delimiter :  string
            delimiter between item objects

    """
    def __init__(self, path, delimiter="."):
        self.path = path
        self.delimiter = delimiter
        if path:
            self.path_parsed = path.split(delimiter)
        else:
            self.path_parsed = None

    def get_top_item(self):
        """
        Returns
        -------
        Item : 'Item'
                a name of item on top of ItemPath
        """
        return self.path_parsed[0]

    def get_next_item_path(self):
        """
        Returns
        -------
        item_path : 'ItemPath'
                path to a nested items from the  item.
        """
        next_part = self.path_parsed[1:]
        if next_part:
            return ItemPath(self.delimiter.join(next_part))

    def __str__(self):
        return self.path



import numpy as np
from numba import njit
from scipy import linalg
import logging

from .linalg.lapack.lapack_python import lapack_solve_triangular, lapack_cholesky

class BaseMethod:
    def __init__(self, numerous_solver, **options):
        raise NotImplementedError
    def get_solver_state(self, *args, **kwargs):
        raise NotImplementedError


class RK45(BaseMethod):
    def __init__(self, numerous_solver, **options):

        submethod=options['submethod']
        profile = numerous_solver.numba_compiled_solver

        def comp(fun):
            if profile:
                return njit(fun)
            else:
                # return options['lp'](fun)
                return fun

        if submethod == None:
            submethod = 'RKDP45'

        if submethod == 'RKF45':

            c = np.zeros(6)
            c[0] = 0
            c[1] = 1 / 4
            c[2] = 3 / 8
            c[3] = 12 / 13
            c[4] = 1
            c[5] = 1 / 2
            a = np.zeros((6, 5))
            b = np.zeros((2, 6))
            a[1, 0] = 1 / 4
            a[2, 0] = 3 / 32
            a[2, 1] = 9 / 32
            a[3, 0] = 1932 / 2197
            a[3, 1] = -7200 / 2197
            a[3, 2] = 7296 / 2197
            a[4, 0] = 439 / 216
            a[4, 1] = -8
            a[4, 2] = 3680 / 513
            a[4, 3] = -845 / 4104
            a[5, 0] = -8 / 27
            a[5, 1] = 2
            a[5, 2] = -3544 / 2565
            a[5, 3] = 1859 / 4104
            a[5, 4] = -11 / 40

            b[1, 0] = 16 / 135
            b[1, 1] = 0
            b[1, 2] = 6656 / 12825
            b[1, 3] = 28561 / 56430
            b[1, 4] = -9 / 50
            b[1, 5] = 2 / 55

            b[0, 0] = 25 / 216
            b[0, 1] = 0
            b[0, 2] = 1408 / 2565
            b[0, 3] = 2197 / 4104
            b[0, 4] = -1 / 5
            b[0, 5] = 0
            self.order = 4
            self.rk_steps = 5
            self.e_order = 5
        elif submethod == 'RKDP45':
            c=np.zeros(7)
            c[0] = 0
            c[1] = 1/5
            c[2] = 3/10
            c[3] = 4/5
            c[4] = 8/9
            c[5] = 1
            c[6] = 1

            a = np.zeros((6,6))
            a[1:] = [1/5, 0, 0, 0, 0 ,0]
            a[2:] = [3/40, 9/40, 0, 0, 0, 0]
            a[3:] = [44/45, -56/15, 32/9, 0, 0, 0]
            a[4:] = [19372/6561, -25360/2187, 64448/6561, -212/729, 0, 0]
            a[5:] = [9017/3168, -355/33, 46732/5247, 49/176, -5103/18656, 0]

            b = np.zeros((2,7))
            b[0:] = [35/384, 0, 500/1113, 125/192, -2187/6784, 11/84, 0]
            b[1:] = [5179/57600, 0, 7571/16695, 393/640, -92097/339200, 187/2100, 1/40]
            self.e_order = 4
            self.order = 5
            self.rk_steps = 5

        else:
            raise Exception("incorrect submethod specified")

        self.f0 = np.ascontiguousarray(numerous_solver.f0)
        self.a = a
        self.b = b
        self.c = c

        self.error_exponent = (-1 / (self.e_order + 1))
        self.max_factor = options.get('max_factor', 10)
        self.atol = options.get('atol', 1e-3)
        self.rtol = options.get('rtol', 1e-3)

        @comp
        def Rk45(nm, t, dt, y, _not_used1, _not_used2, _solve_state):

            c = _solve_state[0]
            a = _solve_state[1]
            b = _solve_state[2]
            max_factor = _solve_state[3]
            atol = _solve_state[4]
            rtol = _solve_state[5]
            f0 = _solve_state[6]
            rk_steps = _solve_state[7]
            order = _solve_state[8]
            error_exponent = _solve_state[9]
            last_step = _solve_state[10]
            step_info = 1

            converged = False

            tnew = t+dt

            if len(y) == 0:
                return tnew, y, True, 1, _solve_state, max_factor

            k = np.zeros((rk_steps+2, len(y)))
            k[0,:] = f0*dt

            for i in range(1,rk_steps+1):
                dy = np.dot(k[:i].T, a[i,:i])
                k[i,:] = dt*nm.func(t+c[i]*dt, y+dy)

            ynew = y + np.dot(k[0:order+2].T, b[0,:])
            fnew = nm.func(tnew, ynew)  # can possibly save one call here...
            k[-1, :] = dt*fnew

            ye = y + np.dot(k[0:order+2].T, b[1,:])
            scale = atol + np.maximum(np.abs(y), np.abs(ynew)) * rtol

            e = (ynew-ye)

            e_norm = np.linalg.norm(e/scale)/ (len(e)**0.5)

            if e_norm < 1:
                if e_norm == 0:
                    factor = max_factor
                else:
                    factor = min(max_factor,
                                 0.95 * e_norm ** error_exponent)

                if not last_step:
                    factor = min(1, factor)

                f0=fnew
                converged = True
            else:
                factor = max(0.2,
                             0.95 * e_norm ** error_exponent)

            _new_solve_state = (c, a, b, max_factor, atol, rtol, np.ascontiguousarray(f0), rk_steps, order, error_exponent, converged)

            return tnew, ynew, converged, step_info, _new_solve_state, factor

        self.step_func = Rk45

    def get_solver_state(self, *args, **kwargs):
        return (self.c, self.a, self.b, self.max_factor, self.atol, self.rtol, self.f0, self.rk_steps, self.order,
                self.error_exponent, True)




class LevenbergMarquardt(BaseMethod):
    def __init__(self, numerous_solver, **options):
        self.order = 5

        l_init = options.get('l', 1)
        self.l_init = l_init
        # l = options.get('l', 1)
        # Get options
        inner_itermax = options.get('inner_itermax', 4)
        S_tol = options.get('S_tol', 1e-8)
        p_tol = options.get('p_tol', 1e-8)
        grad_tol = options.get('grad_tol', 1e-8)
        jac_update_iter = options.get('jac_update_iter', 10)
        jac_update_res_change = options.get('jac_update_res_change', 10)
        update_jacobian_ = options.get('update_jacobian', True)
        abs_tol = options.get('atol', 0.001)
        rel_tol = options.get('rtol', 0.001)
        self.longer = options.get('longer', 1.2)
        self.shorter = options.get('longer', 0.8)
        lp = options.get('lp', None)
        profile = numerous_solver.numba_compiled_solver
        # if lp is not None:
        #     profile = True

        eps = np.finfo(1.0).eps
        newton_tol = max(10 * eps / rel_tol, min(0.03, rel_tol ** 0.5))

        s = options.get('s', 0)
        jacobian_stepsize = options.get('jacobian_stepsize', )

        af = np.array([1, 2 / 3, 6 / 11, 12 / 25, 60 / 137], dtype=np.float)
        a = np.array([[-1, 0, 0, 0, 0], [1 / 3, -4 / 3, 0, 0, 0], [-2 / 11, 9 / 11, -18 / 11, 0, 0],
                      [3 / 25, -16 / 25, 36 / 25, -48 / 25, 0],
                      [-12 / 137, 75 / 137, -200 / 137, 300 / 137, -300 / 137]], dtype=np.float)


        def comp(fun):
            if profile:
                return njit(fun)
            else:
                # return options['lp'](fun)
                return fun


        @comp
        def calc_residual(r):
            Stemp = np.sum(r ** 2)
            S = Stemp / len(r)
            return S

        @comp
        def guess_init(yold, order, last_f, dt):
            _sum = np.zeros_like(yold[0, :])
            for i in range(order):
                _sum += a[order - 1][i] * yold[i, :]
            return -_sum + af[order - 1] * last_f * dt

        @comp
        def sparsejacobian(get_f, get_f_ix, __internal_state, t, y, s, dt, jacobian_stepsize):
            # Limits the number of equation calls by assuming a sparse jacobian matrix, for example in finite element
            num_eq_vars = len(y)
            jac = np.zeros((num_eq_vars, num_eq_vars))
            np.fill_diagonal(jac, 1)
            h = jacobian_stepsize

            f = get_f(t, y, __internal_state)
            y_perm = np.copy(y)

            for i in range(len(y)):
                col_m = max(0, i - s)
                col_p = min(num_eq_vars - 1, i + s)
                idx = [k for k in range(col_m, col_p + 1)]

                for j in idx:
                    y_perm[j] += h
                    f_perm = get_f_ix(t, y_perm, __internal_state, i, j)
                    y_perm[j] -= h
                    diff = (f_perm - f[i]) / h
                    jac[i][j] += -dt * diff

            return jac

        @comp
        def nonesensejac(get_f, get_f_ix, __internal_state, t, y, _, dt, jacobian_stepsize):
            jac = np.zeros((len(y), len(y)))
            np.fill_diagonal(jac, 1)
            return jac

        @comp
        def fulljacobian(get_f, get_f_ix, __internal_state, t, y, _, dt, jacobian_stepsize):
            num_eq_vars = len(y)
            jac = np.zeros((num_eq_vars, num_eq_vars))
            np.fill_diagonal(jac, 1)

            h = jacobian_stepsize

            f = get_f(t, y, __internal_state)

            for i in range(num_eq_vars):

                for j in range(num_eq_vars):
                    yperm = np.copy(y)

                    yperm[j] += h

                    # f = self.get_f_ix(t, np.array(y, dtype=float), i, j)

                    f_h = get_f(t, yperm, __internal_state)

                    # f_h = get_f_ix(t, yperm, __internal_state, i, j)

                    diff = (f_h[i] - f[i]) / h
                    # print(diff)
                    jac[j][i] += -dt * diff

            return jac

        # @comp
        @comp
        def levenberg_marquardt_inner(nm, t, dt, yinit, yold, jacT, L, order):
            # ll=l
            ll = 0
            stat = 0
            # first estimate of next step

            # TODO: give last known derivative value, instead of calculating again
            y = yinit.copy()

            _iter = 0

            scale = abs_tol + rel_tol * np.abs(y)

            x = np.zeros_like(y)
            d = np.zeros_like(x)
            rtest = np.zeros_like(y)

            x_norm_old = None

            converged = False

            while True:
                _iter += 1

                r, f = nm.get_g(t + dt, yold, y, dt, order , a, af)
                b = jacT @ r

                # x = solve_triangular(L[0] + ll * DD, -b) # deprecated
                # x = solve_triangular(L[0], -b) # From Cholesky decomposition
                # x = np.linalg.solve(L[1], -(L[0].T @ b)) # From QR decomposition - using numpy
                # x = np.linalg.solve(L[0] @ L[1], -b) # Ax = b - using Numpy
                # x = solve_qr_fun(L[0], L[1], -b) # From decomposition using own algorithms
                x = lapack_solve_triangular(L, -b, len(b))

                d += x
                y += x

                # Check if residual is decreasing
                rate = np.linalg.norm(b, ord=2)
                if rate < newton_tol:
                    converged = True
                    break

                if _iter > inner_itermax:
                    stat = -2
                    break

            return converged, rate, y, d, rtest, ll, stat, f

        @comp
        def levenberg_marquardt(nm, t, dt, y, yold, order, _solve_state):
            n = len(y)
            ynew = np.zeros_like(y)

            J = _solve_state[0]
            jacT = _solve_state[1]
            update_jacobian = _solve_state[2]
            last_f = _solve_state[3]
            L = _solve_state[4]
            updated = _solve_state[5]
            update_L = _solve_state[6]
            jac_updates = _solve_state[7]
            longer = _solve_state[8]
            shorter = _solve_state[9]

            converged = False
            y = guess_init(yold, order, last_f, dt)

            while not converged:

                # We need a smarter way to update the jacobian...
                if update_jacobian:
                    # print("update")
                    J = nm.vectorizedfulljacobian(t + dt, y, dt)

                    update_jacobian = False
                    updated = True
                    update_L = True
                    jac_updates += 1
                    # print(jac_updates, t, dt)

                if update_L:
                    jac = np.identity(n) - af[order - 1] * dt * J
                    jacT = jac.T
                    jacjacT = jac.T @ jac
                    L = lapack_cholesky(76, n, jacjacT)
                    update_L = False

                converged, S, ynew, d, r, lnew, inner_stat, f = \
                    levenberg_marquardt_inner(nm, t, dt, y, yold, jacT, L, order)

                if not converged:
                    if not updated:
                        update_jacobian = True
                    else:
                        update_L = True
                        break
                if converged:
                    updated = False
                    last_f = f

                # info = {'outer_stat': outer_stat, 'inner_stat': inner_stat, 'p_conv': p_conv, 'S': S,
                #        'grad_conv': grad_conv, 'outer_iter': outer_iter}
                # TODO: look at this dict giving issues
                # _solve_state = (jacT, L, l, converged, last_f)

            _solve_state = (J, jacT, update_jacobian, last_f, L, updated, update_L, jac_updates, longer, shorter)
            t += dt
            # print(t)
            if not converged:
                factor = shorter
            else:
                factor = longer

            return t, ynew, converged, update_jacobian, _solve_state, factor

        if profile:
            self.lp = options.get('lp')
            self.lp.add_function(levenberg_marquardt)
            self.lp.add_function(levenberg_marquardt_inner)
            # self.lp.add_function(nm.vectorizedfulljacobian)

        self.step_func = levenberg_marquardt

    def get_solver_state(self, n):

        state = (np.ascontiguousarray(np.zeros((n, n))),
                 np.ascontiguousarray(np.zeros((n, n))),
                 True,
                 np.ascontiguousarray(np.zeros(n)),
                 np.ascontiguousarray(np.zeros((n, n))),
                 False,
                 True,
                 0,
                 self.longer,
                 self.shorter,
                 )

        return state



from numerous.engine.system.connector import Connector
from numerous.engine.system.item import Item
from numerous.engine.system.namespace import _BindingVariable


class ConnectorItem(Item, Connector):
    """
    Item that can be used as a connector.
    """

    def __init__(self, tag):
        super(ConnectorItem, self).__init__(tag=tag)

    def create_namespace(self, namespace_name):
        """
        Creating a namespace in item and all bindings.

        Parameters
        ----------
        namespace_name : string
            Name of a `VariableNamespace`

        Returns
        -------
        new_namespace : `VariableNamespace`
            Empty namespace with given name
        """
        new_namespace = super().create_namespace(namespace_name)
        super()._create_shadow_namespace(namespace_name)
        return new_namespace

    def __bind_mappings(self, binding, binded_item):
        for ns in self.registered_namespaces:
            for f_var in ns.variables:
                if f_var.mapping:
                    b_fvar=f_var.mapping
                    if isinstance(b_fvar, _BindingVariable):
                        if b_fvar.namespace.binding.name == binding.name:
                            bv = binded_item.registered_namespaces[ns.tag].get_variable(
                                b_fvar.detailed_description.tag)
                            f_var.mapping = bv

    def bind(self, **kwargs):
        """
        Method to bind item to the bindings in current item. Biding items creating all mappings that
        a linked to Binding.

        Parameters
        ----------
        **kwargs : `Item`
            items to bind in form, binding_name = Item

        """
        for key, value in kwargs.items():
            if key in self.bindings.keys():

                self.bindings[key].add_binding(value)
                self.__bind_mappings(self.bindings[key], value)
            else:
                ValueError("Binding {} is not exist in item {}".format(key, self.tag))


class ConnectorTwoWay(ConnectorItem):
    """
        Special case of  a connector item with 2 predefined bindings.

    """

    def __init__(self, tag, side1_name='side1', side2_name='side2'):
        super(ConnectorTwoWay, self).__init__(tag=tag)
        self.binding_names = [side1_name, side2_name]
        for bn in self.binding_names:
            self.create_binding(bn)

class ConnectorMultiWay(ConnectorItem):
    """
        Special case of  a connector item with 2 predefined bindings.

    """

    def __init__(self, tag, binding_names=[]):
        super(ConnectorMultiWay, self).__init__(tag=tag)
        self.binding_names = binding_names
        for bn in self.binding_names:
            self.create_binding(bn)

class ConnectorOneWay(ConnectorItem):
    """
        Special case of  a connector item with 2 predefined bindings.

    """

    def __init__(self, tag, side1_name='side1'):
        super(ConnectorOneWay, self).__init__(tag=tag)
        self.binding_names = [side1_name]
        for bn in self.binding_names:
            self.create_binding(bn)

class ConnectorMultiWay(ConnectorItem):
    """
        Special case of  a connector item with n predefined bindings.

    """

    def __init__(self, tag, binding_names=[]):
        super(ConnectorMultiWay, self).__init__(tag=tag)

        self.binding_names = binding_names
        for bn in self.binding_names:
            self.create_binding(bn)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


with open('requirements.txt', "r") as requirements:
    install_requires = [s.strip() for s in requirements]

setuptools.setup(
    name="numerous-engine",
    version="0.2.3",
    author='Artem Chupryna, EnergyMachines ApS',
    author_email='artem.chupryna@energymachines.com',
    description="Numerous  - an object-oriented modelling and simulation engine.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fossilfree/numerous",
    packages=setuptools.find_namespace_packages(include=['numerous.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    python_requires='>=3.10',
)



from numerous.engine.variables import VariableType, SetOfVariables
from numerous.engine.model.graph_representation.utils import EdgeType
from numerous.engine.model.utils import NodeTypes
from numerous.utils.string_utils import d_u

from .graph import Graph, Node, Edge


class TemporaryVar():
    def __init__(self, id, svi, tag, set_var, set_var_ix):
        self.temporary_variable = True
        self.id = id
        self.tag = tag + id
        self.scope_var_id = svi.id
        self.set_var = set_var
        self.set_var_ix = set_var_ix
        self.value = svi.value
        self.type = VariableType.TMP_PARAMETER
        self.path = svi.path
        self.item = svi.item

    def get_path_dot(self):
        return self.tag


##should be inherited from SetOfVariables
class TemporarySetVar:

    def __init__(self, id, set_var):
        self.id = id
        self.tmp_vars = []
        self.type = VariableType.TMP_PARAMETER_SET
        self.set_var = set_var
        self.size = set_var.size
        self.variables = self.set_var.variables

    def get_size(self):
        return len(self.tmp_vars)

    def get_var_by_idx(self, i):
        return next(var for var in self.tmp_vars if var.set_var_ix == i)


class SumCount:
    def __init__(self):
        self.count = -1

    def get_sum(self):
        self.count += 1
        return f"sum_{self.count}"


new_sum = SumCount().get_sum


class MappingsGraph(Graph):

    def __init__(self, preallocate_items=1000):
        super().__init__(preallocate_items)
        self.vars_assignments = {}
        self.vars_mappings = {}
        self.vars_assignments_mappings = {}

    def variables(self):
        for n in self.node_map.values():
            if self.get(n, 'node_type') == NodeTypes.VAR:
                yield n

    def remove_chains(self):
        for target in self.variables():
            # Get target
            target_edges_indcs_edge_type_target, target_edges_edge_type_target = self.get_edges_for_node_filter(
                end_node=target, attr='e_type', val=EdgeType.TARGET)
            target_edges_indcs_edge_type_mapping, target_edges_edge_type_mapping = self.get_edges_for_node_filter(
                end_node=target, attr='e_type', val=EdgeType.MAPPING)
            target_edges_indcs = target_edges_indcs_edge_type_target + target_edges_indcs_edge_type_mapping
            target_edges = target_edges_edge_type_target + target_edges_edge_type_mapping
            for edge, edge_ix in zip(target_edges, target_edges_indcs):

                if not target in self.vars_assignments:
                    self.vars_assignments[target] = []
                    self.vars_assignments_mappings[target] = []

                if self.edges_c[edge_ix].e_type == EdgeType.MAPPING:
                    self.vars_mappings[target] = (edge[0], self.edges_c[edge_ix].mappings)
                    self.remove_edge(edge_ix)

                self.vars_assignments[target].append(edge[0])
                self.vars_assignments_mappings[target].append(self.edges_c[edge_ix].mappings)

        for target in self.variables():
            target_edges_indcs, target_edges = self.get_edges_for_node_filter(end_node=target, attr='e_type',
                                                                              val=[EdgeType.TARGET, EdgeType.MAPPING])
            if target in self.vars_assignments and len(self.vars_assignments[target]) > 1:
                for edge_ix in target_edges_indcs:
                    self.remove_edge(edge_ix)

    def create_assignments(self, variables):
        from tqdm import tqdm
        temp_variables = {}
        for ii, n in tqdm(enumerate(self.get_where_node_attr('node_type', NodeTypes.EQUATION)),
                          desc="Creating assigments"):
            for i, e in self.get_edges_for_node(start_node=n):
                va = e[1].copy()
                if va in self.vars_assignments and len(self.vars_assignments[va]) > 1:
                    # Make new temp var
                    sv = self.get(e[1], 'scope_var')
                    tmp_key = sv.tag + str(self.key_map[va]) + '_tmp'
                    tmp_label = sv.tag + str(self.key_map[va]) + '_label' + '_tmp'
                    # Create fake scope variables for tmp setvar
                    fake_sv = {}
                    svf = None
                    if isinstance(sv, SetOfVariables):
                        tmp_var_counter = 0
                        tsv = TemporarySetVar(tmp_key, sv)
                        for svi in sv.variables.values():
                            tmp_var_counter += 1
                            svf = TemporaryVar('tmp_var_' + str(tmp_var_counter), svi,
                                               svi.tag + '_' + str(sv.id), svi.set_var, svi.set_var_ix)
                            tsv.tmp_vars.append(svf)
                        fake_sv[tsv.id] = tsv
                    else:
                        svf = TemporaryVar(d_u(tmp_key), sv, tmp_key, None, None)
                        fake_sv[d_u(svf.get_path_dot())] = svf

                    temp_variables.update(fake_sv)

                    tmp = self.add_node(Node(key=tmp_key, node_type=NodeTypes.TMP, name=tmp_key,
                                             file='sum', label=tmp_label, ln=0, scope_var=svf), ignore_existing=False)
                    # Add temp var to Equation target

                    self.add_edge(Edge(n, tmp, e_type=EdgeType.TARGET, arg_local=self.edges_c[i[0]].arg_local))
                    # Add temp var in var assignments

                    self.vars_assignments_mappings[va][(nix := self.vars_assignments[va].index(n))] = ':'
                    self.vars_assignments[va][nix] = tmp
        return temp_variables

    def add_mappings(self):
        for a, vals in self.vars_assignments.items():
            if len(vals) > 1:
                ns = new_sum()
                nsn = self.add_node(Node(key=ns, node_type=NodeTypes.SUM, name=ns, file='sum',
                                         label=ns,
                                         ln=0, ast_type=None))
                self.add_edge(Edge(start=nsn, end=a, e_type=EdgeType.TARGET))
                for v, mappings in zip(vals, self.vars_assignments_mappings[a]):
                    self.add_edge(Edge(start=v, end=nsn, e_type=EdgeType.VALUE, mappings=mappings))

            elif a in self.vars_mappings:
                ns = new_sum()
                nsn = self.add_node(Node(key=ns, node_type=NodeTypes.SUM, name=ns, file='sum',
                                         label=ns,
                                         ln=0, ast_type=None))
                self.add_edge(Edge(start=nsn, end=a, e_type=EdgeType.TARGET))

                self.add_edge(Edge(start=self.vars_mappings[a][0], end=nsn,
                                   e_type=EdgeType.VALUE, mappings=self.vars_mappings[a][1]))

    @classmethod
    def from_graph(cls, eg):
        eg.__class__ = cls
        eg.vars_assignments = {}
        eg.vars_mappings = {}
        eg.vars_assignments_mappings = {}
        return eg



import logging
import time
from collections import namedtuple
from enum import IntEnum, unique

import numpy as np
from numba import njit

from numerous.engine.simulation.solvers.base_solver import BaseSolver
from .solver_methods import BaseMethod, RK45

Info = namedtuple('Info', ['status', 'event_id', 'step_info', 'dt', 't', 'y', 'order_', 'roller', 'solve_state'])


@unique
class SolveStatus(IntEnum):
    Running = 0
    Finished = 1


@unique
class SolveEvent(IntEnum):
    NoneEvent = 0
    Historian = 1
    ExternalDataUpdate = 2


class Numerous_solver(BaseSolver):

    def __init__(self, time_, delta_t, model, numba_model, num_inner, max_event_steps, y0, numba_compiled_solver,
                 events,
                 event_directions,
                 timestamp_events,
                 **kwargs):
        super().__init__()

        self.events = events[0][0]
        self.event_directions = event_directions
        self.actions = events[1][0]
        self.timestamps = timestamp_events[1]

        self.timestamps_actions = timestamp_events[0][0]
        # events value
        self.g = self.events(time_[0], y0)

        self.time = time_
        self.model = model
        self.num_inner = num_inner
        self.delta_t = delta_t
        self.numba_model = numba_model
        self.diff_function = numba_model.func
        self.numba_compiled_solver = numba_compiled_solver
        self.number_of_events = len(self.g)

        self.f0 = numba_model.func(time_[0], y0)
        self.max_event_steps = max_event_steps
        self.options = kwargs
        self.info = None

        eps = np.finfo(1.0).eps
        odesolver_options = {'longer': kwargs.get('longer', 1.2), 'shorter': kwargs.get('shorter', 0.8),
                             'min_step': kwargs.get('min_step', 10 * eps), 'strict_eval': True,
                             'max_step': kwargs.get('max_step', np.inf), 'first_step': kwargs.get('first_step', None),
                             'atol': kwargs.get('atol', 1e-6), 'rtol': kwargs.get('rtol', 1e-3),
                             'outer_itermax': kwargs.get('outer_itermax', 20),
                             'submethod': kwargs.get('submethod', None),
                             }

        self.method_options = odesolver_options
        try:
            self.method = eval(kwargs.get('method', 'RK45'))
            self._method = self.method(self, **self.method_options)
            assert issubclass(self.method, BaseMethod), f"{self.method} is not a BaseMethod"
        except Exception as e:
            raise e

        self.y0 = y0

        # Generate the solver
        if numba_compiled_solver:
            self._non_compiled_solve = njit(self.generate_solver())
            self._solve = self.compile_solver()
        else:
            self._solve = self.generate_solver()

    def generate_solver(self):
        def _solve(numba_model, _solve_state, initial_step, order, order_, roller, strict_eval, outer_itermax,
                   min_step, max_step, step_integrate_, events, actions, g, number_of_events, event_directions,
                   timestamps, timestamp_actions,
                   t0=0.0, t_end=1000.0, t_eval=np.linspace(0.0, 1000.0, 100)):
            # Init t to t0
            imax = 100
            step_info = 0
            t = t0
            dt = initial_step
            y = numba_model.get_states()
            if y.shape[0] == 0:
                for t in t_eval[1:]:
                    numba_model.func(t, y)
                    numba_model.historian_update(t)
                    numba_model.map_external_data(t)
                return Info(status=SolveStatus.Finished, event_id=SolveEvent.NoneEvent, step_info=step_info,
                            dt=dt, t=t, y=y, order_=order_, roller=roller, solve_state=_solve_state)
            t_start = t
            t_previous = t0
            y_previous = np.copy(y)

            len_y = numba_model.get_states().shape[0]

            def add_ring_buffer(t_, y_, rb, o):

                if o == order:
                    y_temp = rb[2][:, :]
                    t_temp = rb[1]
                    rb[1][0:order - 1] = t_temp[1:order]
                    rb[2][0:order - 1, :] = y_temp[1:order, :]

                o = min(o + 1, order)
                rb[1][o - 1] = t_
                rb[2][o - 1, :] = y_

                return o

            def get_order_y(rb, order):
                y = rb[2][0:order, :]
                return y

            # Define event derivatives, values and event guess times
            # edt = np.ones(max(1, n_events)) * tol + 2.0
            # et = np.zeros(n_events) if n_events > 0 else np.ones(1)
            te_array = np.zeros(2)
            # 0 index is used to keep next time step defined by solver
            te_array[0] = t
            # 1 index is used to keep next time to eval/save the solution
            ix_eval = 1
            te_array[1] = t_eval[ix_eval] + dt if strict_eval else np.inf
            t_next_eval = t_eval[ix_eval]

            outer_iter = outer_itermax

            terminate = False
            step_converged = True
            event_trigger = False
            event_ix = -1
            t_event = t
            y_event = y
            dt_last = -1
            j_i = 0
            progress_c = t_eval.shape[0]
            while not terminate:
                # updated events time estimates
                # # time acceleration
                if not step_converged:

                    if min_step > dt:
                        raise ValueError('dt shortened below min_dt')
                    decrease = True
                    te_array[0] = t_previous + dt
                elif step_converged and not event_trigger:
                    dt = min(max_step, dt)
                    decrease = False

                    te_array[0] = t + dt
                else:  # event
                    te_array[0] = t_event

                # Determine new test - it should be the smallest value requested by events, eval, step
                t_new_test = np.min(te_array)

                # Determine if rollback is needed
                # check if t_new_test is forward
                # no need to roll back if time-step is decreased, as it's not a failed step
                if (t_new_test < t) or (not step_converged):
                    # Need to roll back!
                    t_start = t_previous
                    y = y_previous

                    if t_new_test < t_start:
                        # t_new_test = t_rollback
                        # TODO: make more specific error raising here!
                        raise ValueError('Cannot go back longer than rollback point!')
                else:
                    # Since we didnt roll back we can update t_start and rollback
                    # Check if we should update history at t eval
                    if t_next_eval <= (t + 10 * np.finfo(1.0).eps):
                        j_i += 1
                        p_size = 100
                        x = int(p_size * j_i / progress_c)
                        if not step_converged:
                            print("step not converged, but historian updated")
                        numba_model.historian_update(t)
                        if strict_eval:
                            te_array[1] = t_next_eval = t_eval[ix_eval + 1] if ix_eval + 1 < len(t_eval) else t_eval[-1]
                        else:
                            t_next_eval = t_eval[ix_eval + 1] if ix_eval + 1 < len(t_eval) else t_eval[-1]
                        ix_eval += 1
                        dt = initial_step
                        te_array[0] = t + dt

                    t_start = t
                    t_new_test = np.min(te_array)
                    if t >= t_end:
                        break

                    order_ = add_ring_buffer(t, y, roller, order_)

                dt_ = min([t_next_eval - t_start, t_new_test - t_start])
                # solve from start to new test by calling the step function

                t, y, step_converged, step_info, _solve_state, factor = step_integrate_(numba_model,
                                                                                        t_start,
                                                                                        dt_, y,
                                                                                        get_order_y(roller, order_),
                                                                                        order_,
                                                                                        _solve_state)

                dt *= factor
                event_trigger = False
                t_events = np.zeros(number_of_events) + t
                y_events = np.zeros((len(y), number_of_events))

                def sol(t, t_r, y_r):
                    yi = np.zeros(len(y))
                    tv = np.append(roller[1][0:order], t_r)
                    yv = np.append(roller[2][0:order], y_r)
                    yv = yv.reshape(order + 1, len(y)).T
                    for i, yvi in enumerate(yv):
                        yi[i] = np.interp(t, tv, yvi)
                    return yi

                def check_event(event_fun, ix, t_previous, y_previous, t, y, t_next_eval):
                    t_l = t_previous
                    y_l = y_previous
                    e_l = event_fun(t_l, y_l)[ix]
                    t_r = t
                    y_r = y
                    e_r = event_fun(t_r, y_r)[ix]
                    status = 0
                    if np.sign(e_l) == np.sign(e_r):
                        return status, t, y
                    i = 0
                    t_m = (t_l + t_r) / 2
                    y_m = sol(t_m, t, y)

                    while status == 0:  # bisection method
                        e_m = event_fun(t_m, y_m)[ix]
                        if np.sign(e_l) != np.sign(e_m):
                            t_r = t_m
                        elif np.sign(e_r) != np.sign(e_m):
                            t_l = t_m
                        if abs(e_m) < 1e-6:
                            status = 1
                        if i > imax:
                            status = -1
                        t_m = (t_l + t_r) / 2
                        y_m = sol(t_m, t, y)

                    return status, t_m, y_m

                if step_converged:
                    g_new = events(t, y)
                    up = (g <= 0) & (g_new >= 0) & (event_directions == 1)
                    down = (g >= 0) & (g_new <= 0) & (event_directions == -1)
                    g = g_new

                    for ix in np.concatenate((np.argwhere(up), np.argwhere(down))):
                        status, t_event, y_event = check_event(events, ix[0],
                                                               t_previous, y_previous, t, y, t_next_eval)
                        t_events[ix[0]] = t_event
                        y_events[:, ix[0]] = y_event

                if min(t_events) < t:
                    event_trigger = True
                    event_ix = np.argmin(t_events)
                    t_event = t_events[event_ix]
                    y_event = y_events[:, event_ix]
                    g = events(t_event, y_event)

                if not event_trigger and step_converged:
                    y_previous = y
                    t_previous = t
                    for event_ix, timestamps_ in enumerate(timestamps):
                        for timestamp in timestamps_:
                            if abs(timestamp - t) < 1e-6:
                                numba_model.set_states(y)
                                modified_variables = timestamp_actions(t, numba_model.read_variables(), event_ix)
                                modified_mask = (modified_variables != numba_model.read_variables())
                                for idx in np.argwhere(modified_mask):
                                    numba_model.write_variables(modified_variables[idx[0]], idx[0])
                                y_previous = numba_model.get_states()

                if event_trigger:
                    numba_model.set_states(y_event)
                    modified_variables = actions(t_event, numba_model.read_variables(), event_ix)
                    modified_mask = (modified_variables != numba_model.read_variables())
                    for idx in np.argwhere(modified_mask):
                        numba_model.write_variables(modified_variables[idx[0]], idx[0])
                    y_previous = numba_model.get_states()
                    t_previous = t_event
                    numba_model.historian_update(t_event)

                if step_converged:
                    numba_model.map_external_data(t)
                    if numba_model.is_store_required():
                        return Info(status=SolveStatus.Running, event_id=SolveEvent.Historian, step_info=step_info,
                                    dt=dt, t=t, y=np.ascontiguousarray(y), order_=order_, roller=roller,
                                    solve_state=_solve_state)
                    if numba_model.is_external_data_update_needed(t):
                        return Info(status=SolveStatus.Running, event_id=SolveEvent.ExternalDataUpdate,
                                    step_info=step_info, dt=dt, t=t, y=np.ascontiguousarray(y), order_=order_,
                                    roller=roller, solve_state=_solve_state)

            return Info(status=SolveStatus.Finished, event_id=SolveEvent.NoneEvent, step_info=step_info,
                        dt=dt, t=t, y=np.ascontiguousarray(y), order_=order_, roller=roller, solve_state=_solve_state)

        return _solve

    def compile_solver(self):

        logging.info("Compiling Numerous Solver")
        generation_start = time.time()

        argtypes = []

        max_step = self.method_options.get('max_step')
        min_step = self.method_options.get('min_step')

        strict_eval = self.method_options.get('strict_eval', True)
        outer_itermax = self.method_options.get('outer_itermax', 20)

        order = self._method.order
        initial_step = min_step

        step_integrate_ = self._method.step_func
        roller = self._init_roller(order)
        order_ = 0

        args = (self.numba_model,
                self._method.get_solver_state(len(self.y0)), initial_step,
                order, order_, roller, strict_eval, outer_itermax, min_step,
                max_step, step_integrate_,
                self.events,
                self.actions,
                self.g,
                self.number_of_events,
                self.event_directions,
                self.timestamps,
                self.timestamps_actions,
                self.time[0],
                self.time[-1],
                self.time)
        for a in args:
            argtypes.append(self._non_compiled_solve.typeof_pyval(a))
        # Return the solver function

        _solve = self._non_compiled_solve.compile(tuple(argtypes))

        generation_finish = time.time()
        logging.info(f"Solver compiled, compilation time: {generation_finish - generation_start}")

        return _solve

    def set_state_vector(self, states_as_vector):
        self.y0 = states_as_vector

    def select_initial_step(self, nm, t0, y0, direction, order, rtol, atol):
        """Taken from scipy select initial step part of the ode solver package. Slightly modified


        Empirically select a good initial step.

        The algorithm is described in [1]_.

        Parameters
        ----------
        nm : Numbamodel - contains the function handles to call the RHS

        t0 : float
            Initial value of the independent variable.
        y0 : ndarray, shape (n,)
            Initial value of the dependent variable.
        direction : float
            Integration direction.
        order : float
            Error estimator order. It means that the error controlled by the
            algorithm is proportional to ``step_size ** (order + 1)`.
        rtol : float
            Desired relative tolerance.
        atol : float
            Desired absolute tolerance.

        Returns
        -------
        h_abs : float
            Absolute value of the suggested initial step.

        References
        ----------
        .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
               Equations I: Nonstiff Problems", Sec. II.4.
        """

        f0 = self.f0

        if y0.size == 0:
            return np.inf

        scale = atol + np.abs(y0) * rtol
        d0 = np.linalg.norm(y0 / scale)
        d1 = np.linalg.norm(f0 / scale)
        if d0 < 1e-5 or d1 < 1e-5:
            h0 = 1e-6
        else:
            h0 = 0.01 * d0 / d1

        y1 = y0 + h0 * direction * f0
        f1 = nm.func(t0 + h0 * direction, y1)
        d2 = np.linalg.norm((f1 - f0) / scale) / h0

        if d1 <= 1e-15 and d2 <= 1e-15:
            h1 = max(1e-6, h0 * 1e-3)
        else:
            h1 = (0.01 / max(d1, d2)) ** (1 / (order + 1))

        # Restore states in numba model
        nm.func(t0, y0)

        return min(100 * h0, h1)

    def _init_roller(self, order):
        n = order + 2
        rb0 = np.zeros((n, len(self.y0)))
        roller = (n, np.zeros(n), rb0)
        return roller

    def solve(self):
        """
        solve the model.

        Returns
        -------
        Solution : 'OdeSolution'
                returns the most recent OdeSolution from scipy

        """
        self.result_status = "Success"
        self.sol = None
        logging.info('Solve started')
        # try:
        t_start = self.time[0]
        t_end = self.time[-1]

        # Call the solver
        from copy import deepcopy
        y0 = deepcopy(self.y0)

        state = deepcopy(self.y0)

        # Set options

        max_step = self.method_options.get('max_step')
        min_step = self.method_options.get('min_step')
        rtol = self.method_options.get('rtol')
        atol = self.method_options.get('atol')

        order = self._method.order

        initial_step = self.select_initial_step(self.numba_model, t_start, y0, 1, order - 1, rtol,
                                                atol)  # np.min([100000000*min_step, max_step])

        strict_eval = self.method_options.get('strict_eval')
        outer_itermax = self.method_options.get('outer_itermax')

        step_integrate_ = self._method.step_func

        # figure out solve_state init
        solve_state = self._method.get_solver_state(len(y0))

        roller = self._init_roller(order)
        order_ = 0

        info = self._solve(self.numba_model,
                           solve_state, initial_step, order, order_, roller, strict_eval, outer_itermax, min_step,
                           max_step, step_integrate_, self.events, self.actions, self.g,
                           self.number_of_events, self.event_directions, self.timestamps,
                           self.timestamps_actions, t_start, t_end, self.time)

        while info.status == SolveStatus.Running:
            if info.event_id == 1:
                self.model.create_historian_df()
                self.numba_model.historian_reinit()
            if info.event_id == 2:
                is_external_data = self.model.external_mappings.load_new_external_data_batch(info.t)
                external_mappings_numpy = self.model.external_mappings.external_mappings_numpy
                external_mappings_time = self.model.external_mappings.external_mappings_time
                self.numba_model.is_external_data = is_external_data
                self.numba_model.update_external_data(external_mappings_numpy, external_mappings_time)
            time_idx =np.argmax((self.time - info.t)>0)
            info = self._solve(self.numba_model,
                               solve_state, info.dt, order, strict_eval, outer_itermax, min_step,
                               max_step, step_integrate_, self.events, self.actions, self.g,
                               self.number_of_events, self.event_directions, self.timestamps, self.timestamps_actions,
                               info.t, t_end, self.time[time_idx:])

        logging.info("Solve finished")
        return self.sol, self.result_status

    def solver_step(self, t, delta_t=None):

        solve_state = self._method.get_solver_state(len(self.y0))
        t_start = t

        if delta_t is None:
            delta_t = self.delta_t

        strict_eval = self.method_options.get('strict_eval')
        outer_itermax = self.method_options.get('outer_itermax')

        max_step = self.method_options.get('max_step')
        min_step = self.method_options.get('min_step')
        rtol = self.method_options.get('rtol')
        atol = self.method_options.get('atol')
        order = self._method.order
        if self.info is not None:
            dt = self.info.dt  # internal solver step size
            order_ = self.info.order_
            roller = self.info.roller
            solve_state = self.info.solve_state

            assert self.info.t == t_start, f"solver time {self.info.t} does not match external time " \
                                           f"{t_start}"
        else:
            roller = self._init_roller(order)
            order_ = 0

            dt = self.select_initial_step(self.numba_model, t_start, self.y0, 1, order - 1, rtol,
                                          atol)

        t_end = t_start + delta_t

        time_span = np.linspace(t_start, t_end, 2)

        step_integrate_ = self._method.step_func

        info = self._solve(self.numba_model,
                           solve_state, dt, order, order_, roller, strict_eval, outer_itermax, min_step,
                           max_step, step_integrate_, self.events, self.actions, self.g,
                           self.number_of_events, self.event_directions, self.timestamps, self.timestamps_actions,
                           t_start, t_end, time_span)

        if info.event_id == 1:
            self.model.store_history(self.numba_model.historian_data)
            self.numba_model.historian_reinit()
        if info.event_id == 2:
            is_external_data = self.model.external_mappings.load_new_external_data_batch(info.t)
            external_mappings_numpy = self.model.external_mappings.external_mappings_numpy
            external_mappings_time = self.model.external_mappings.external_mappings_time
            self.numba_model.is_external_data = is_external_data
            self.numba_model.update_external_data(external_mappings_numpy, external_mappings_time)

        self.info = info

        return t_end, self.info.t

    def register_endstep(self, __end_step):
        self.__end_step = __end_step



from abc import ABC
from typing import Tuple

import pandas as pd
from pandas import DataFrame
from pandas.errors import EmptyDataError


class DataLoader(ABC):

    def __init__(self):
        self.chunksize = 1000

    def load(self, df_id: str, t: int):
        pass


class LocalDataLoader(DataLoader):

    def __init__(self, chunksize=1000):
        self.chunksize = chunksize
        self.is_chunks = False
        if self.chunksize is not None:
            self.is_chunks = True

    def load(self, df_id: str, t: int) -> DataFrame:
        if self.is_chunks:
            return pd.read_csv(df_id, header=0, skiprows=t, chunksize=self.chunksize).get_chunk()
        else:
            return pd.read_csv(df_id, header=0, skiprows=t, chunksize=self.chunksize)


class InMemoryDataLoader(DataLoader):

    def __init__(self, df):
        self.df = df

    def load(self, df_id: str, t: int) -> DataFrame:
        return self.df




import logging
from datetime import datetime
import time
import numpy as np
from numerous.engine.simulation.solvers.base_solver import SolverType
from numerous.engine.simulation.solvers.ivp_solver.ivp_solver import IVP_solver
from numerous.engine.simulation.solvers.numerous_solver.numerous_solver import Numerous_solver


class Simulation:
    """
    Class that wraps simulation solver. currently only solve_ivp.

    Attributes
    ----------
          time :  ndarray
               Not unique tag that will be used in reports or printed output.
          solver: :class:`BaseSolver`
               Instance of differential equations solver that will be used for the model.
          delta_t :  float
               timestep.
          callbacks :  list
               Not unique tag that will be used in reports or printed output.
          model :
               Not unique tag that will be used in reports or printed output.
    """

    def __init__(self, model, solver_type=SolverType.SOLVER_IVP, t_start=0, t_stop=20000, num=1000, num_inner=1,
                 max_event_steps=100,

                 start_datetime=datetime.now(), **kwargs):
        """
            Creating a namespace.

            Parameters
            ----------
            tag : string
                Name of a `VariableNamespace`

            Returns
            -------
            new_namespace : `VariableNamespace`
                Empty namespace with given name
        """

        time_, delta_t = np.linspace(t_start, t_stop, num + 1, retstep=True)
        self.callbacks = []
        self.time = time_
        self.model = model

        def __end_step(solver, y, t, events_action, event_id=None):
            """

            """
            solver.y0 = y.flatten()

            if event_id is not None:
                self.model.update_local_variables()
                ## slow code
                list_var = [v.value for v in self.model.path_to_variable.values()]
                events_action[event_id](t, list_var)
                for i, var in enumerate(self.model.path_to_variable.values()):
                    var.value = list_var[i]
                self.model.update_all_variables()
                solver.y0 = self.model.states_as_vector

            else:
                solver.numba_model.historian_update(t)
                solver.numba_model.map_external_data(t)

                if solver.numba_model.is_store_required():
                    self.model.store_history(numba_model.historian_data)
                    solver.numba_model.historian_reinit()
                #
                if solver.numba_model.is_external_data_update_needed(t):
                    solver.numba_model.is_external_data = self.model.external_mappings.load_new_external_data_batch(t)
                    if solver.numba_model.is_external_data:
                        solver.numba_model.update_external_data(self.model.external_mappings.external_mappings_numpy,
                                                                self.model.external_mappings.external_mappings_time)

        self.end_step = __end_step
        logging.info("Generating Numba Model")
        generation_start = time.time()
        logging.info(f'Number of steps: {len(self.time)}')
        numba_model = model.generate_compiled_model(t_start, len(self.time))

        generation_finish = time.time()
        logging.info(f"Numba model generation finished, generation time: {generation_finish - generation_start}")
        if solver_type.value == SolverType.SOLVER_IVP.value:
            event_function, _ = model.generate_event_condition_ast(False)
            action_function = model.generate_event_action_ast(model.events, False)
            timestamp_action_function = model.generate_event_action_ast(model.timestamp_events, False)
            timestamps = np.array([np.array(event.timestamps) for event in model.timestamp_events])
            self.solver = IVP_solver(time_, delta_t, model, numba_model,
                                     num_inner, max_event_steps, self.model.states_as_vector,
                                     events=(event_function, action_function),
                                     timestamp_events=(timestamp_action_function, timestamps),
                                     **kwargs)

        if solver_type.value == SolverType.NUMEROUS.value:
            event_function, event_directions = model.generate_event_condition_ast(True)
            action_function = model.generate_event_action_ast(model.events, True)
            if len(model.timestamp_events) == 0:
                model.generate_mock_timestamp_event()
            timestamp_action_function = model.generate_event_action_ast(model.timestamp_events, True)
            timestamps = np.array([np.array(event.timestamps) for event in model.timestamp_events])
            self.solver = Numerous_solver(time_, delta_t, model, numba_model,
                                          num_inner, max_event_steps, self.model.states_as_vector,
                                          numba_compiled_solver=model.use_llvm,
                                          events=(event_function, action_function), event_directions=event_directions,
                                          timestamp_events=(timestamp_action_function, timestamps),
                                          **kwargs)

        self.solver.register_endstep(__end_step)

        self.start_datetime = start_datetime
        self.info = model.info["Solver"]
        self.info["Number of Equation Calls"] = 0

        logging.info("Compiling Numba equations and initializing historian")
        compilation_start = time.time()
        numba_model.func(t_start, numba_model.get_states())
        numba_model.historian_update(t_start)
        compilation_finished = time.time()
        logging.info(
            f"Numba equations compiled, historian initizalized, compilation time: {compilation_finished - compilation_start}")

        self.compiled_model = numba_model

    def solve(self):
        self.reset()

        sol, self.result_status = self.solver.solve()

        self.info.update({"Solving status": self.result_status})
        self.complete()
        return sol

    def reset(self):
        self.__init_step()
        self.model.numba_model.historian_reinit()

        self.end_step(self.solver, self.model.numba_model.get_states(), 0, [])

    def step(self, dt):
        try:
            stop = self.solver.solver_step(dt)
        except Exception as e:
            raise e

        return stop

    def complete(self):

        list(map(lambda x: x.restore_variables_from_numba(self.solver.numba_model,
                                                          self.model.path_variables), self.model.callbacks))
        self.model.create_historian_df()

    def step_solve(self, t, step_size):
        try:
            t, results_status = self.solver.solver_step(t, step_size)

            return t, results_status
        except Exception as e:
            raise e

    def __init_step(self):
        pass
